from contextlib import suppress
import curses

import visidata

from visidata import EscapeException, ExpectedException, clipdraw, Sheet, VisiData, BaseSheet
from visidata import vd, colors, dispwidth, ColorAttr, clipstr_start
from visidata import AttrDict


vd.theme_option('color_edit_unfocused', '238 on 110', 'display color for unfocused input in form')
vd.theme_option('color_edit_cell', '233 on 110', 'cell color to use when editing cell')
vd.theme_option('disp_edit_fill', '_', 'edit field fill character')
vd.theme_option('disp_unprintable', '·', 'substitute character for unprintables')
vd.theme_option('mouse_interval', 1, 'max time between press/release for click (ms)', sheettype=None)

vd.disp_help = 0  # current page of help shown
vd._help_sidebars = []  # list of (help:str|HelpPane, title:str)


class AcceptInput(Exception):
    '*args[0]* is the input to be accepted'

vd._injectedInput = None  # for vd.injectInput


@VisiData.api
def injectInput(vd, x):
    'Use *x* as input to next command.'
    assert vd._injectedInput is None, vd._injectedInput
    vd._injectedInput = x


@VisiData.api
def getCommandInput(vd):
    if vd._injectedInput is not None:
        r = vd._injectedInput
        vd._injectedInput = None
        return r

    return vd.getLastArgs()


@BaseSheet.after
def execCommand(sheet, longname, *args, **kwargs):
    if vd._injectedInput is not None:
        vd.debug(f'{longname} did not consume input "{vd._injectedInput}"')
        vd._injectedInput = None


def acceptThenFunc(*longnames):
    def _acceptthen(v, i):
        for longname in longnames:
            vd.queueCommand(longname)
        raise AcceptInput(v)
    return _acceptthen

# editline helpers

class EnableCursor:
    def __enter__(self):
        with suppress(curses.error):
            curses.mousemask(0)
            curses.curs_set(1)

    def __exit__(self, exc_type, exc_val, tb):
        with suppress(curses.error):
            curses.curs_set(0)
            if vd.options.mouse_interval:
                curses.mousemask(curses.MOUSE_ALL if hasattr(curses, "MOUSE_ALL") else 0xffffffff)
            else:
                curses.mousemask(0)


def until_get_wch(scr):
    'Ignores get_wch timeouts'
    ret = None
    while not ret:
        try:
            ret = vd.get_wch(scr)
        except curses.error:
            pass

    if isinstance(ret, int):
        return chr(ret)
    return ret


def splice(v:str, i:int, s:str):
    'Insert `s` into string `v` at `i` (such that v[i] == s[0]).'
    return v if i < 0 else v[:i] + s + v[i:]


@VisiData.api
def drawInputHelp(vd, scr):
    if not scr or not vd.cursesEnabled:
        return

    sheet = vd.activeSheet
    if not sheet:
        return

    vd.drawSidebar(scr, sheet)


def clean_printable(s):
    'Escape unprintable characters.'
    return ''.join(c if c.isprintable() else vd.options.disp_unprintable for c in str(s))


def delchar(s, i, remove=1):
    'Delete `remove` characters from str `s` beginning at position `i`.'
    return s if i < 0 else s[:i] + s[i+remove:]

def find_word(s, a, b, incr):
        '''Return first index of word boundary in s[a:b], going forward if incr is +1 and backward if incr is -1.'''
        if not s: return 0
        a = min(max(a, 0), len(s)-1)
        b = min(max(b, 0), len(s)-1)

        if incr < 0:
            while not s[b].isalnum() and b >= a:  # first skip non-word chars
                b += incr
            while s[b].isalnum() and b >= a:
                b += incr
            return min(max(b, -1), len(s))
        else:
            while s[a].isalnum() and a < b:       # first skip word chars
                a += incr
            while not s[a].isalnum() and a < b:   # then skip non-word chars
                a += incr
            return min(max(a, 0), len(s))

class InputWidget:
    def __init__(self,
                 value:str='',
                 i=0,
                 display=True,
                 history=[],
                 completer=lambda text,idx: None,
                 options=None,
                 fillchar=''):
        '''
            - value: starting value
            - i: starting index into value
            - display: False to not display input (for sensitive input, e.g. a password)
            - history: list of strings; earliest entry first.
            - completer: func(value:str, idx:int) takes the current value and tab completion index, and returns a string if there is a completion available, or None if not.
            - options: sheet.options; defaults to vd.options.
        '''
        options = options or vd.options

        self.orig_value = value
        self.first_action = (i == 0)  # whether this would be the 'first action'; if so, clear text on input

        # display theme
        self.fillchar  = fillchar or options.disp_edit_fill
        self.truncchar = options.disp_truncator
        self.display = display  # if False, obscure before displaying

        # main state
        self.value = self.orig_value  # value under edit
        self.current_i = i
        self.insert_mode = True

        # history state
        self.history = history
        self.hist_idx = None
        self.prev_val = None

        # completion state
        self.comps_idx = -1
        self.completer_func = completer
        self.former_i = None
        self.just_completed = False

    def editline(self, scr, y, x, w, attr=ColorAttr(), updater=lambda val:None, bindings={}, clear=True) -> str:
        'If *clear* is True, clear whole editing area before displaying.'
        with EnableCursor():
            while True:
                if len(vd.pendingKeys) <= 3:  #speed up paste of long strings by skipping redraws
                    vd.drawSheet(scr, vd.activeSheet)
                    updater(self.value)
                    vd.drawInputHelp(scr)

                self.draw(scr, y, x, w, attr, clear=clear)
                ch = vd.getkeystroke(scr)
                if ch in bindings:
                    self.value, self.current_i = bindings[ch](self.value, self.current_i)
                else:
                    if self.handle_key(ch, scr):
                        return self.value


    def draw(self, scr, y, x, w, attr=ColorAttr(), clear=True):
        i = self.current_i  # the onscreen offset within the field where v[i] is displayed
        trunch = self.truncchar
        tr_w = dispwidth(trunch)
        fill_w = dispwidth(self.fillchar)

        def _calc_display(dispval, i):
            '''Return a formatted substring of *dispval* that fills the on-screen width *w*.'''
            if i == len(dispval): # add a fillchar so the user perceives room to type
                dispval += self.fillchar
            dw = dispwidth(dispval)
            if dw <= w:  # entire value fits
                dispval += self.fillchar*(w-dw)
                return dispval, i
            if w <= tr_w: # column is too narrow to hold a left and right truncation
                return trunch, 0

            dw = dispwidth(dispval[i:])
            if dw + tr_w <= w and dw <= w//2: #cursor is within half-colwidth of end
                #truncate the left and show the end
                frag, n = clipstr_start(dispval, w-tr_w)
                offset = len(dispval) - i
                dispval = ' '*(w-tr_w - n) + trunch + frag
                i = len(dispval) - offset
                return dispval, i

            # the remaining cases need the right side truncated, after the new dispval is returned
            dw = dispwidth(dispval[:i+1])
            if dw + tr_w <= w and dispwidth(dispval[:i]) <= w//2: #cursor is within half-colwidth of start
                #truncate the right, and show the string start
                pass
            else: # truncate left and right sides
                # Place the cursor at the midpoint of the available colwidth
                left_w = (w - 2*tr_w)//2
                # calculate the fragment to the left of the cursor
                l_frag, n = clipstr_start(dispval[:i], left_w)
                dispval = ' '*(left_w-n) + trunch + l_frag + dispval[i:]
                i = left_w-n + len(trunch) + len(l_frag)
            return dispval, i

        if self.display:
            dispval = clean_printable(self.value)
        else:
            dispval = '*' * len(self.value)
        dispval, i = _calc_display(dispval, i)

        #clipdraw will truncate the right side of dispval with trunch as needed
        clipdraw(scr, y, x, dispval, attr, w, clear=clear, literal=True)
        if x+w < scr.getmaxyx()[1]:
            #draw a space to indicate that the user can scroll right of the cell's final char
            clipdraw(scr, y, x+w, ' ', attr, 1, clear=False, literal=True)
        if scr:
            prew = dispwidth(dispval[:i])
            if x+prew < scr.getmaxyx()[1]: #move cursor back to where the user is editing
                scr.move(y, x+prew)

    def handle_key(self, ch:str, scr) -> bool:
        'Return True to accept current input.  Raise EscapeException on Ctrl+C, Ctrl+Q, or ESC.'
        i = self.current_i
        v = self.value

        if ch == '':                               return False
        elif ch == 'KEY_IC':                       self.insert_mode = not self.insert_mode
        elif ch == '^A' or ch == 'KEY_HOME':       i = 0
        elif ch == '^B' or ch == 'KEY_LEFT':       i -= 1
        elif ch in ('^C', '^Q', '^['):             raise EscapeException(ch)
        elif ch == '^D' or ch == 'KEY_DC':         v = delchar(v, i)
        elif ch == '^E' or ch == 'KEY_END':        i = len(v)
        elif ch == '^F' or ch == 'KEY_RIGHT':      i += 1
        elif ch == '^G':
            vd.cycleSidebar()
            return False # not considered a first keypress
        elif ch in ('^H', 'KEY_BACKSPACE', '^?'):  i -= 1; v = delchar(v, i)
        elif ch == '^I':                           v, i = self.completion(v, i, +1)
        elif ch == 'KEY_BTAB':                     v, i = self.completion(v, i, -1)
        elif ch in ['^J', '^M']:                   return True # ENTER to accept value
        elif ch == '^K':                           v = v[:i]  # ^Kill to end-of-line
        elif ch == '^N':
            c = ''
            while not c:
                c = vd.getkeystroke(scr)
            c = vd.prettykeys(c)
            i += len(c)
            v += c
        elif ch == '^O':
            edit_v = vd.launchExternalEditor(v)
            if self.value == edit_v:
                # leave cell unmodified when the editor exits with no change
                raise EscapeException(ch)
            else:
                self.value = edit_v
                return True
        elif ch == '^R':                           v = self.orig_value  # ^Reload initial value
        elif ch == '^T':                           v = delchar(splice(v, i-2, v[i-1:i]), i)  # swap chars
        elif ch == '^U':                           v = v[i:]; i = 0  # clear to beginning
        elif ch == '^V':                           v = splice(v, i, until_get_wch(scr)); i += 1  # literal character
        elif ch == '^W':                           j = find_word(v, 0, i-1, -1); v = v[:j+1] + v[i:]; i = j+1  # erase word
        elif ch in ('KEY_DC5','kDC5','kDC3'):      j = find_word(v, i, len(v), +1); v = v[:i] + v[j+1:]  # erase word forward
        elif ch == '^Y':                           v = splice(v, i, str(vd.memory.clipval))
        elif ch == '^Z':                           vd.suspend()
        # CTRL+arrow
        elif ch == 'kLFT5':                        i = find_word(v, 0, i-1, -1)+1;  # word left
        elif ch == 'kRIT5':                        i = find_word(v, i, len(v)-1, +1);  # word right
        elif ch == 'kUP5':                         pass
        elif ch == 'kDN5':                         pass
        elif self.history and ch == 'KEY_UP':    v, i = self.prev_history(v, i)
        elif self.history and ch == 'KEY_DOWN':  v, i = self.next_history(v, i)
        elif len(ch) > 1:                          pass
        else:
            if self.first_action:
                v = ''
            if self.insert_mode:
                v = splice(v, i, ch)
            else:
                v = v[:i] + ch + v[i+1:]

            i += 1

        if i < 0: i = 0
        # v may have a non-str type with no len()
        v = str(v)
        if i > len(v): i = len(v)
        self.current_i = i
        self.value = v
        self.first_action = False
        self.reset_completion()
        return False

    def completion(self, v, i, state_incr):
        self.just_completed = True
        self.comps_idx += state_incr

        if self.former_i is None:
            self.former_i = i
        try:
            r = self.completer_func(v[:self.former_i], self.comps_idx)
        except Exception as e:
            # raise  # beep/flash; how to report exception?
            return v, i

        if not r:
            # beep/flash to indicate no matches?
            return v, i

        v = r + v[i:]
        return v, len(v)

    def reset_completion(self):
        if self.just_completed:
            self.just_completed = False
        else:
            self.former_i = None
            self.comps_idx = -1

    def prev_history(self, v, i):
        if self.hist_idx is None:
            self.hist_idx = len(self.history)
            self.prev_val = v
        if self.hist_idx > 0:
            self.hist_idx -= 1
            v = self.history[self.hist_idx]
        i = len(str(v))
        return v, i

    def next_history(self, v, i):
        if self.hist_idx is None:
            return v, i
        elif self.hist_idx < len(self.history)-1:
            self.hist_idx += 1
            v = self.history[self.hist_idx]
        else:
            v = self.prev_val
            self.hist_idx = None
        i = len(str(v))
        return v, i


@VisiData.api
def editText(vd, y, x, w, attr=ColorAttr(), value='',
             help='',
             updater=lambda val: None, bindings={},
             display=True, record=True, clear=True, **kwargs):
    '''Invoke modal single-line editor at (*y*, *x*) for *w* terminal chars. Use *display* is False for sensitive input like passphrases.  If *record* is True, get input from the cmdlog in batch mode, and save input to the cmdlog if *display* is also True. Return new value as string. Callers should handle curses.error, which will be raised if the terminal is resized during the edit, in a way that moves the editor coordinates offscreen.'''
    v = None
    if record and vd.cmdlog:
        v = vd.getCommandInput()

    if v is None:
        if vd.options.batch:
            return ''

        if vd.activeSheet._scr is None:
            raise Exception('active sheet does not have a screen')

        if value is None:
            value = ''

        try:
            widget = InputWidget(value=str(value), display=display, **kwargs)

            with vd.AddedHelp(vd.getHelpPane('input', module='visidata'), 'Input Keystrokes Help', 'inputkeys'), \
                 vd.AddedHelp(help, 'Input Field Help', 'inputfield'):
                v = widget.editline(vd.activeSheet._scr, y, x, w, attr=attr, updater=updater, bindings=bindings, clear=clear)
        except AcceptInput as e:
            v = e.args[0]

        if vd.cursesEnabled:
            # clear keyboard buffer to neutralize multi-line pastes (issue#585)
            curses.flushinp()

    if display:
        if record and vd.cmdlog:
            vd.setLastArgs(v)

    if value:
        if isinstance(value, (int, float)) and v[-1] == '%':  #2082
            pct = float(v[:-1])
            v = pct*value/100

        # convert back to type of original value
        v = type(value)(v)

    return v

@VisiData.api
def inputsingle(vd, prompt, record=True):
    'Display prompt and return single character of user input.'
    sheet = vd.activeSheet

    v = None
    if record and vd.cmdlog:
        v = vd.getCommandInput()

    if v is not None:
        return v

    y = sheet.windowHeight-1
    w = sheet.windowWidth
    rstatuslen = vd.drawRightStatus(sheet._scr, sheet)
    promptlen = clipdraw(sheet._scr, y, 0, prompt, 0, w=w-rstatuslen-1)
    sheet._scr.move(y, w-promptlen-rstatuslen-2)

    while not v:
        v = vd.getkeystroke(sheet._scr)

    if record and vd.cmdlog:
        vd.setLastArgs(v)

    return v

@VisiData.api
def inputMultiple(vd, updater=lambda val: None, record=True, **kwargs):
    'A simple form, where each input is an entry in `kwargs`, with the key being the key in the returned dict, and the value being a dictionary of kwargs to the singular input().'
    sheet = vd.activeSheet
    scr = sheet._scr

    previnput = vd.getCommandInput()
    if previnput is not None:
        ret = None
        if isinstance(previnput, str):
            if previnput.startswith('{'):
                ret = json.loads(previnput)
            else:
                ret = {k:v.get('value', '') for k,v in kwargs.items()}
                primekey = list(ret.keys())[0]
                ret[primekey] = previnput

        if isinstance(previnput, dict):
            ret = previnput

        if ret:
            if record and vd.cmdlog:
                vd.setLastArgs(ret)
            return ret

        assert False, type(previnput)

    maxw = sheet.windowWidth//2
    attr = colors.color_edit_unfocused

    keys = list(kwargs.keys())
    cur_input_key = keys[0]

    if scr:
        scr.erase()

    for i, (k, v) in enumerate(kwargs.items()):
        v['dy'] = i
        v['w'] = maxw-dispwidth(v.get('prompt'))

    class ChangeInput(Exception):
        pass

    def change_input(offset):
        def _throw(v, i):
            if scr:
                scr.erase()
            raise ChangeInput(v, offset)
        return _throw

    def _drawPrompt(val):
        for k, v in kwargs.items():
            #recalculate y to adjust for screen resizes during input()
            y = sheet.windowHeight-v.get('dy')-1
            maxw = min(sheet.windowWidth-1, max(dispwidth(v.get('prompt')), dispwidth(str(v.get('value', '')))))
            promptlen = clipdraw(scr, y, 0, v.get('prompt'), attr, w=maxw)  #1947
            promptlen = clipdraw(scr, y, promptlen, v.get('value', ''),  attr, w=maxw)

        return updater(val)

    while True:
        try:
            input_kwargs = kwargs[cur_input_key]
            input_kwargs['value'] = vd.input(**input_kwargs,
                                             attr=colors.color_edit_cell,
                                             updater=_drawPrompt,
                                             record=False,
                                             bindings={
                'KEY_BTAB': change_input(-1),
                '^I':       change_input(+1),
                'KEY_SR':   change_input(-1),
                'KEY_SF':   change_input(+1),
                'kUP':      change_input(-1),
                'kDN':      change_input(+1),
            })
            break
        except ChangeInput as e:
            input_kwargs['value'] = e.args[0]
            offset = e.args[1]
            i = keys.index(cur_input_key)
            cur_input_key = keys[(i+offset)%len(keys)]

    retargs = {}
    lastargs = {}
    for k, input_kwargs in kwargs.items():
        v = input_kwargs.get('value', '')
        retargs[k] = v

        if input_kwargs.get('record', record):
            if input_kwargs.get('display', True):
                lastargs[k] = v
                vd.addInputHistory(v, input_kwargs.get('type', ''))
    if record:
        if vd.cmdlog and lastargs:
            vd.setLastArgs(lastargs)

    return retargs


@VisiData.api
def input(vd, prompt, type=None, defaultLast=False, history=[], dy=0, attr=None, updater=lambda v: None, **kwargs):
    '''Display *prompt* and return line of user input.

        - *type*: string indicating the type of input to use for history.
        - *history*: list of strings to use for input history.
        - *defaultLast*:  on empty input, if True, return last history item.
        - *display*: pass False to not display input (for sensitive input, e.g. a password), and to also prevent recording input as if *record* is False
        - *record*: pass False to not record input on cmdlog or input history (for sensitive or inconsequential input).
        - *completer*: ``completer(val, idx)`` is called on TAB to get next completed value.
        - *updater*: ``updater(val)`` is called every keypress or timeout.
        - *bindings*: dict of keystroke to func(v, i) that returns updated (v, i)
        - *dy*: number of lines from bottom of pane
        - *attr*: curses attribute for prompt
        - *help*: string to include in help
    '''

    if attr is None:
        attr = ColorAttr()
    sheet = vd.activeSheet
    if not vd.cursesEnabled:
        if kwargs.get('record', True) and vd.cmdlog:
            return vd.getCommandInput()

        if kwargs.get('display', True):
            import builtins
            return builtins.input(prompt)
        else:
            import getpass
            return getpass.getpass(prompt)

    if not history:
        history = list(vd.inputHistory.setdefault(type, {}).keys())

    y = sheet.windowHeight-dy-1
    promptlen = dispwidth(prompt)

    def _drawPrompt(val=''):
        rstatuslen = vd.drawRightStatus(sheet._scr, sheet)
        clipdraw(sheet._scr, y, 0, prompt, attr, w=sheet.windowWidth-rstatuslen-1)
        updater(val)
        return sheet.windowWidth-promptlen-rstatuslen-2

    w = kwargs.pop('w', _drawPrompt())
    restarts = 0
    while restarts < 100:
        #recalculate y to handle resize events
        y = sheet.windowHeight-dy-1
        try:
            ret = vd.editText(y, promptlen, w=w,
                                attr=colors.color_edit_cell,
                                options=vd.options,
                                history=history,
                                updater=_drawPrompt,
                                **kwargs)
            if ret:
                if kwargs.get('record', True) and kwargs.get('display', True):
                    vd.addInputHistory(ret, type=type)
            elif defaultLast:
                history or vd.fail("no previous input")
                ret = history[-1]

            return ret
        except curses.error:
            vd.warning('restarting input due to resize')
            restarts += 1
    # if it keeps happening, it's probably not resize events, so give some debug output
    vd.error(f'aborting input:  y={y}, w={w}, windowHeight={sheet.windowHeight}, windowWidth={sheet.windowWidth}')


@VisiData.api
def confirm(vd, prompt, exc=EscapeException):
    'Display *prompt* on status line and demand input that starts with "Y" or "y" to proceed.  Raise *exc* otherwise.  Return True.'
    if vd.options.batch:
        return vd.fail('cannot confirm in batch mode: ' + prompt)

    yn = vd.input(prompt, value='no', record=False)[:1]
    if not yn or yn not in 'Yy':
        msg = 'disconfirmed: ' + prompt
        if exc:
            raise exc(msg)
        vd.warning(msg)
        return False
    return True


class CompleteKey:
    def __init__(self, items):
        self.items = items

    def __call__(self, val, state):
        opts = [x for x in self.items if x.startswith(val)]
        return opts[state%len(opts)] if opts else val


@Sheet.api
def editCell(self, vcolidx=None, rowidx=None, value=None, **kwargs):
    '''Call vd.editText for the cell at (*rowidx*, *vcolidx*).  Return the new value, properly typed.
       - *vcolidx*: numeric index into ``self.availCols``. When None, use current column.
       - *rowidx*: numeric index into ``self.rows``.  If negative, indicates the column name in the header.
       - *value*: if given, the starting input; otherwise the starting input is the cell value or column name as appropriate.
       - *kwargs*: passthrough args to ``vd.editText``.
       '''

    if vcolidx is None:
        vcolidx = self.cursorVisibleColIndex
    x, w = self._visibleColLayout.get(vcolidx, (0, 0))

    col = self.availCols[vcolidx]
    if rowidx is None:
        rowidx = self.cursorRowIndex

    if rowidx < 0:  # header
        y = 0
        value = value or col.name
    else:
        y, h = self._rowLayout.get(rowidx, (0, 0))
        value = value or col.getDisplayValue(self.rows[self.cursorRowIndex])

    bindings={
        'kUP':        acceptThenFunc('go-up', 'rename-col' if rowidx < 0 else 'edit-cell'),
        'KEY_SR':     acceptThenFunc('go-up', 'rename-col' if rowidx < 0 else 'edit-cell'),
        'kDN':        acceptThenFunc('go-down', 'rename-col' if rowidx < 0 else 'edit-cell'),
        'KEY_SF':     acceptThenFunc('go-down', 'rename-col' if rowidx < 0 else 'edit-cell'),
        'KEY_SRIGHT': acceptThenFunc('go-right', 'rename-col' if rowidx < 0 else 'edit-cell'),
        'KEY_SLEFT':  acceptThenFunc('go-left', 'rename-col' if rowidx < 0 else 'edit-cell'),
        '^I':         acceptThenFunc('go-right', 'rename-col' if rowidx < 0 else 'edit-cell'),
        'KEY_BTAB':   acceptThenFunc('go-left', 'rename-col' if rowidx < 0 else 'edit-cell'),
    }

    if vcolidx == self.nVisibleCols-1 or vcolidx >= self.nCols-1:
        bindings['^I'] = acceptThenFunc('go-down', 'go-leftmost', 'edit-cell')

    if vcolidx <= 0:
        bindings['KEY_BTAB'] = acceptThenFunc('go-up', 'go-rightmost', 'edit-cell')

    # update local bindings with kwargs.bindings instead of the inverse, to preserve kwargs.bindings for caller
    bindings.update(kwargs.get('bindings', {}))
    kwargs['bindings'] = bindings

    editargs = dict(value=value, options=self.options)

    editargs.update(kwargs)  # update with user-specified args
    try:
        r = vd.editText(y, x, w, attr=colors.color_edit_cell, **editargs)
    except curses.error:
        vd.fail(f'aborting edit due to resize')

    if rowidx >= 0:  # if not header
        r = col.type(r)  # convert input to column type, let exceptions be raised

    return r


vd.addGlobals(CompleteKey=CompleteKey, AcceptInput=AcceptInput, InputWidget=InputWidget)



import collections
import itertools
from copy import copy, deepcopy
import textwrap

from visidata import VisiData, Extensible, globalCommand, ColumnAttr, ColumnItem, vd, ENTER, EscapeException, drawcache, drawcache_property, LazyChainMap, asyncthread, ExpectedException
from visidata import (options, Column, namedlist, SettableColumn, AttrDict, DisplayWrapper,
TypedExceptionWrapper, BaseSheet, UNLOADED, wrapply,
clipdraw, clipdraw_chunks, ColorAttr, update_attr, colors, undoAttrFunc, vlen, dispwidth)
import visidata
from visidata.utils import colname_letters


vd.activePane = 1   # pane numbering starts at 1; pane 0 means active pane


vd.option('name_joiner', '_', 'string to join sheet or column names')
vd.option('value_joiner', ' ', 'string to join display values')
vd.option('max_rows', 1_000_000_000, 'number of rows to load from source')

vd.option('disp_wrap_max_lines', 3, 'max lines for multiline view')
vd.option('disp_wrap_break_long_words', False, 'break words longer than column width in multiline')
vd.option('disp_wrap_replace_whitespace', False, 'replace whitespace with spaces in multiline')
vd.option('disp_wrap_placeholder', '…', 'multiline string to indicate truncation')
vd.option('disp_multiline_focus', True, 'only multiline cursor row')
vd.option('color_multiline_bottom', '', 'color of bottom line of multiline rows')  #2715
vd.option('color_aggregator', 'bold 255 white on 234 black', 'color of aggregator summary on bottom row')


@drawcache
def _splitcell(sheet, s, width=0, maxheight=1):
    height = max(maxheight, sheet.options.disp_wrap_max_lines or 0)
    if width <= 0 or height <= 0:
        return [s]

    wrap_kwargs = sheet.options.getall('disp_wrap_')
    wrap_kwargs['max_lines'] = height

    ret = []
    for attr, text in s:
        for line in textwrap.wrap(text, width=width, **wrap_kwargs):
            if len(ret) >= maxheight:
                ret[-1][0][1] += ' ' + line
                break
            else:
                ret.append([[attr, line]])
    return ret

disp_column_fill = ' ' # pad chars before column value

class Colorizer:
    '''higher precedence color overrides lower; all non-color attributes combine.
       coloropt is the color option name (like 'color_error').
       func(sheet,col,row,value) should return a true value if coloropt should be applied
       If coloropt is None, func() should return a coloropt (or None) instead'''

    def __init__(self, precedence:int, coloropt:str, func=lambda s,c,r,v: None):
        self.precedence = precedence
        self.coloropt = coloropt
        self._func = func

class RowColorizer(Colorizer):
    def func(self, s, c, r, v):
        return r is not None and self._func(s,c,r,v)

class ColumnColorizer(Colorizer):
    def func(self, s, c, r, v):
        return c is not None and self._func(s,c,r,v)

class CellColorizer(Colorizer):
    def func(self, s, c, r, v):
        return r is not None and c is not None and self._func(s,c,r,v)


class RecursiveExprException(Exception):
    pass

class LazyComputeRow:
    'Calculate column values as needed.'
    def __init__(self, sheet, row, col=None, **kwargs):
        self.row = row
        self.col = col
        self.sheet = sheet
        self.extra = AttrDict(kwargs) # extra bindings
        self._usedcols = set()

        self._lcm.clear()  # reset locals on lcm

    @property
    def _lcm(self):
        lcmobj = self.col or self.sheet
        if not hasattr(lcmobj, '_lcm'):
            lcmobj._lcm = LazyChainMap(self.sheet, self.col, self.extra, *vd.contexts)
        return lcmobj._lcm

    def __iter__(self):
        yield from self.sheet.availColnames
        yield from self._lcm.keys()
        yield 'row'
        yield 'sheet'
        yield 'col'

    def keys(self):
        return list(self.__iter__())

    def __str__(self):
        return str(self.as_dict())

    def as_dict(self):
        return {c.name:self[c.name] for c in self.sheet.visibleCols}

    def __getattr__(self, k):
        return self.__getitem__(k)

    def __getitem__(self, colid):
        try:
            i = self.sheet.availColnames.index(colid)
            c = self.sheet.availCols[i]
            if c is self.col:  # ignore current column
                j = self.sheet.availColnames[i+1:].index(colid)
                c = self.sheet.availCols[i+j+1]

        except ValueError:
            try:
                c = self._lcm[colid]
            except (KeyError, AttributeError) as e:
                if colid == 'sheet': return self.sheet
                elif colid == 'row': return self
                elif colid == '_row': return self.row
                elif colid == 'col': c = self.col
                else:
                    raise KeyError(colid) from e

        if not isinstance(c, Column):  # columns calc in the context of the row of the cell being calc'ed
            return c

        if c in self._usedcols:
            raise RecursiveExprException()

        self._usedcols.add(c)
        ret = c.getTypedValue(self.row)
        self._usedcols.remove(c)
        return ret

class BasicRow(collections.defaultdict):
    def __init__(self, *args):
        collections.defaultdict.__init__(self, lambda: None)
    def __bool__(self):
        return True

class TableSheet(BaseSheet):
    'Base class for sheets with row objects and column views.'
    _rowtype = lambda: BasicRow()
    _coltype = SettableColumn

    rowtype = 'rows'
    guide = '# {sheet.help_title}\n'

    @property
    def help_title(self):
        if isinstance(self.source, visidata.Path):
            return 'Source Table'
        else:
            return 'Table Sheet'

    columns = []  # list of Column
    colorizers = [ # list of Colorizer
        CellColorizer(2, 'color_default_hdr', lambda s,c,r,v: r is None),
        ColumnColorizer(2, 'color_current_col', lambda s,c,r,v: c is s.cursorCol),
        ColumnColorizer(1, 'color_key_col', lambda s,c,r,v: c and c.keycol),
        CellColorizer(0, 'color_default', lambda s,c,r,v: True),
        RowColorizer(1, 'color_error', lambda s,c,r,v: isinstance(r, (Exception, TypedExceptionWrapper))),
        CellColorizer(3, 'color_current_cell', lambda s,c,r,v: c is s.cursorCol and r is s.cursorRow),
        ColumnColorizer(1, 'color_hidden_col', lambda s,c,r,v: c and c.hidden),
    ]
    nKeys = 0  # columns[:nKeys] are key columns
    _ordering = []  # list of (col:Column|str, reverse:bool)

    def __init__(self, *names, rows=UNLOADED, **kwargs):
        super().__init__(*names, rows=rows, **kwargs)
        self.cursorRowIndex = 0  # absolute index of cursor into self.rows
        self.cursorVisibleColIndex = 0  # index of cursor into self.visibleCols

        self._topRowIndex = 0     # cursorRowIndex of topmost row
        self.leftVisibleColIndex = 0    # cursorVisibleColIndex of leftmost column
        self.rightVisibleColIndex = 0

        # as computed during draw()
        self._rowLayout = {}      # [rowidx] -> (y, w)
        self._visibleColLayout = {}      # [vcolidx] -> (x, w)

        # list of all columns in display order
        self.initialCols = kwargs.pop('columns', None) or type(self).columns
        self.colname_ctr = 0
        self.resetCols()

        self._ordering = list(type(self)._ordering)  #2254
        self._colorizers = self.classColorizers
        self.recalc()  # set .sheet on columns and start caches

        self.__dict__.update(kwargs)  # also done earlier in BaseSheet.__init__

    @property
    def topRowIndex(self):
        return self._topRowIndex

    @topRowIndex.setter
    def topRowIndex(self, v):
        self._topRowIndex = v
        self._rowLayout.clear()

    def addColorizer(self, c):
        'Add Colorizer *c* to the list of colorizers for this sheet.'
        self._colorizers.append(c)
        self._colorizers = sorted(self._colorizers, key=lambda x: x.precedence, reverse=True)

    def removeColorizer(self, c):
        'Remove Colorizer *c* from the list of colorizers for this sheet.'
        self._colorizers.remove(c)

    @property
    def classColorizers(self) -> list:
        'List of all colorizers from sheet class hierarchy in precedence order (highest precedence first)'
        # all colorizers must be in the same bucket
        # otherwise, precedence does not get applied properly
        _colorizers = set()

        for b in [self] + list(type(self).superclasses()):
            for c in getattr(b, 'colorizers', []):
                _colorizers.add(c)

        return sorted(_colorizers, key=lambda x: x.precedence, reverse=True)

    def _colorize(self, col, row, value=None) -> ColorAttr:
        'Return ColorAttr for the given colorizers/col/row/value'

        colorstack = []
        for colorizer in self._colorizers:
            try:
                r = colorizer.func(self, col, row, value)
                if r:
                    colorstack.append((colorizer.precedence, colorizer.coloropt if colorizer.coloropt else r))
            except Exception as e:
                vd.exceptionCaught(e)

        return colors.resolve_colors(tuple(colorstack))

    def addRow(self, row, index=None):
        'Insert *row* at *index*, or append at end of rows if *index* is None.'
        if index is None:
            self.rows.append(row)
        else:
            self.rows.insert(index, row)
            if self.cursorRowIndex and self.cursorRowIndex >= index:
                self.cursorRowIndex += 1
        return row

    def newRow(self):
        'Return new blank row compatible with this sheet.  Overridable.'
        return type(self)._rowtype()

    @drawcache_property
    def colsByName(self):
        'Return dict of colname:col'
        # dict comprehension in reverse order so first column with the name is used
        return {col.name:col for col in self.columns[::-1]}

    def column(self, colname):
        'Return first column whose name matches *colname*.'
        return self.colsByName.get(colname) or vd.fail('no column matching "%s"' % colname)

    def recalc(self):
        'Clear caches and set the ``sheet`` attribute on all columns.'
        for c in self.columns:
            c.recalc(self)

    @asyncthread
    def reload(self):
        'Load or reload rows and columns from ``self.source``.  Async.  Override resetCols() or loader() in subclass.'
        with visidata.ScopedSetattr(self, 'loading', True):
            self.resetCols()
            self.beforeLoad()
            try:
                self.loader()
                vd.debug(f'finished loading {self}')
            finally:
                self.afterLoad()

        self.recalc()

    def beforeLoad(self):
        pass

    def resetCols(self):
        'Reset columns to class settings or constructor settings'
        self.columns = []
        for c in self.initialCols:
            self.addColumn(deepcopy(c))
            if c.disp_expert and vd.wantsHelp('nometacols'):
                c.hide()

        self.setKeys(self.columns[:self.nKeys])

    def loader(self):
        'Reset rows and sync load ``source`` via iterload.  Overridable.'
        try:
            for r in self._iterloader():
                pass
        except FileNotFoundError:
            return  # let it be a blank sheet without error

    def _iterloader(self):
        self.rows = []
        with vd.Progress(gerund='loading', total=0):
            max_rows = self.options.max_rows
            for i, r in enumerate(self.iterload()):
                if self.precious and i >= max_rows:
                    break
                self.addRow(r)
                yield r

    def iterload(self):
        'Generate rows from ``self.source``.  Override in subclass.'
        if False:
            yield vd.fail('no iterload for this loader yet')

    def loadStart(self):
        self.loaditer = self._iterloader()

    def loadSome(self):
        if not self.loaditer:
            return False
        try:
            next(self.loaditer)
            return True
        except StopIteration:
            self.loaditer = None
            return False

    def afterLoad(self):
        'hook for after loading has finished.  Overridable (be sure to call super).'
        # if an ordering has been specified, sort the sheet
        if self._ordering:
            vd.sync(self.sort())

    def iterrows(self):
        if self.rows is UNLOADED:
            try:
                self.rows = []
                for row in self.iterload():
                    self.addRow(row)
                    yield row
                return
            except ExpectedException:
                vd.sync(self.reload())

        for row in vd.Progress(self.rows):
            yield row

    def __iter__(self):
        for row in self.iterrows():
            yield LazyComputeRow(self, row)


    def __copy__(self):
        'Copy sheet design but remain unloaded. Deepcopy columns so their attributes (width, type, name) may be adjusted independently of the original.'
        ret = super().__copy__()
        ret.rows = UNLOADED

        ret.columns = []

        col_mapping = {}
        for c in self.columns:
            new_col = copy(c)
            col_mapping[c] = new_col
            ret.addColumn(new_col)

        ret.setKeys([col_mapping[c] for c in self.columns if c.keycol])

        ret._ordering = []
        for sortcol,reverse in self._ordering:
            if isinstance(sortcol, str):
                ret._ordering.append((sortcol,reverse))
            else:
                ret._ordering.append((col_mapping[sortcol],reverse))

        ret.topRowIndex = ret.cursorRowIndex = 0
        return ret

    @property
    def bottomRowIndex(self):
        return self.topRowIndex+self.nScreenRows-1

    @bottomRowIndex.setter
    def bottomRowIndex(self, newidx):
        self._topRowIndex = newidx-self.nScreenRows+1

    @drawcache_property
    def rowHeight(self):
        cols = self.visibleCols
        return max(c.height for c in cols) if cols else 1

    def __deepcopy__(self, memo):
        'same as __copy__'
        ret = self.__copy__()
        memo[id(self)] = ret
        return ret

    def __str__(self):
        return self.name

    def __repr__(self):
        return f'<{type(self).__name__}: {self.name}>'

    @drawcache_property
    def currow(self):
        return LazyComputeRow(self, self.cursorRow, self.cursorCol)

    def evalExpr(self, expr:str, row=None, col=None, **kwargs):
        'eval() expr in the context of (row, col), with extra bindings in kwargs'
        if row is not None:
            # contexts are cached by sheet/rowid for duration of drawcycle
            contexts = vd._evalcontexts.setdefault((self, self.rowid(row), col), LazyComputeRow(self, row, col, **kwargs))
        else:
            contexts = dict(sheet=self)

        return eval(expr, vd.getGlobals(), contexts)

    def rowid(self, row):
        'Return a unique and stable hash of the *row* object.  Must be fast.  Overridable.'
        return id(row)

    @property
    def nScreenRows(self):
        'Number of visible rows at the current window height.'
        n = (self.windowHeight-self.nHeaderRows-self.nFooterRows)
        if self.options.disp_multiline_focus:  # focus multiline mode
            return n-self.rowHeight+1
        return n//self.rowHeight

    @drawcache_property
    def nHeaderRows(self):
        vcols = self.visibleCols
        return max(0, 1, *(len(col.name.split('\n')) for col in vcols))

    @property
    def nFooterRows(self):
        'Number of lines reserved at the bottom, including status line.'
        return len(self.allAggregators) + 1

    @property
    def cursorCol(self):
        'Current Column object.'
        vcols = self.availCols
        return vcols[min(self.cursorVisibleColIndex, len(vcols)-1)] if vcols else None

    @property
    def cursorRow(self):
        'The row object at the row cursor.'
        idx = self.cursorRowIndex
        return self.rows[idx] if self.nRows > idx else None

    @property
    def visibleRows(self):  # onscreen rows
        'List of rows onscreen.'
        return self.rows[self.topRowIndex:self.topRowIndex+self.nScreenRows]

    @drawcache_property
    def visibleCols(self):  # non-hidden cols
        'List of non-hidden columns in display order.'
        return (self.keyCols + [c for c in self.columns if not c.hidden and not c.keycol]) or [Column('', sheet=self)]

    @drawcache_property
    def keyCols(self):
        'List of visible key columns.'
        return sorted([c for c in self.columns if c.keycol and not c.hidden], key=lambda c:c.keycol)

    @drawcache_property
    def availCols(self):
        'List of all available columns, visible columns first.'
        return self.visibleCols + [c for c in self.columns if c.hidden]

    @drawcache_property
    def availColnames(self):
        'List of all available column names, visible columns first.'
        return [c.name for c in self.availCols]

    @property
    def cursorColIndex(self):
        'Index of current column into `Sheet.columns`. Linear search; prefer `cursorCol` or `cursorVisibleColIndex`.'
        try:
            return self.columns.index(self.cursorCol)
        except ValueError:
            return 0

    @property
    def nonKeyVisibleCols(self):
        'List of visible non-key columns.'
        return [c for c in self.columns if not c.hidden and c not in self.keyCols]

    @property
    def keyColNames(self):
        'String of key column names, for SheetsSheet convenience.'
        return ' '.join(c.name for c in self.keyCols)

    @keyColNames.setter
    def keyColNames(self, v):  #2122
        'Set key columns on this sheet to the space-separated list of column names.'
        newkeys = [self.column(colname) for colname in v.split()]
        self.unsetKeys(self.keyCols)
        self.setKeys(newkeys)

    @property
    def cursorCell(self):
        'Displayed value (DisplayWrapper) at current row and column.'
        return self.cursorCol.getCell(self.cursorRow)

    @property
    def cursorDisplay(self):
        'Displayed value (DisplayWrapper.text) at current row and column.'
        return self.cursorCol.getDisplayValue(self.cursorRow)

    @property
    def cursorTypedValue(self):
        'Typed value at current row and column.'
        return self.cursorCol.getTypedValue(self.cursorRow)

    @property
    def cursorValue(self):
        'Raw value at current row and column.'
        return self.cursorCol.getValue(self.cursorRow)

    def getTypedRow(self, rownum):
        return [c.getTypedValue(self.rows[rownum]) for c in self.availCols]

    @property
    def statusLine(self):
        'Position of cursor and bounds of current sheet.'
        rowinfo = 'row %d (%d selected)' % (self.cursorRowIndex, self.nSelectedRows)
        colinfo = 'col %d (%d visible)' % (self.cursorVisibleColIndex, len(self.visibleCols))
        return '%s  %s' % (rowinfo, colinfo)

    @property
    def nRows(self):
        'Number of rows on this sheet.'
        return len(self.rows)

    @property
    def nCols(self):
        'Number of columns on this sheet.'
        return len(self.columns)

    @property
    def nVisibleCols(self):
        'Number of visible columns on this sheet.'
        return len(self.visibleCols)

    def cursorDown(self, n=1):
        'Move cursor down `n` rows (or up if `n` is negative).'
        self.cursorRowIndex += n

    def cursorRight(self, n=1):
        'Move cursor right `n` visible columns (or left if `n` is negative).'
        self.cursorVisibleColIndex += n

    def addColumn(self, *cols, index=None):
        '''Insert all *cols* into columns at *index*, or append to end of columns if *index* is None.
           If *index* is None, columns are being added by loader, instead of by user.
           If added by user, mark sheet as modified.
           Columns added by loader share sheet's defer status.
           Columns added by user are not marked as deferred.
           Return first column.'''
        if not cols:
            vd.warning('no columns to add')
            return

        if index is not None:
            self.setModified()

        for i, col in enumerate(cols):
            col.name = self.maybeClean(col.name)
            col.defer = self.defer

            vd.addUndo(self.columns.remove, col)
            idx = len(self.columns) if index is None else index
            col.recalc(self)
            self.columns.insert(idx+i, col)

        # statements after addColumn in the same command may want to use these cached properties
        Sheet.keyCols.fget.cache_clear()
        Sheet.visibleCols.fget.cache_clear()
        Sheet.availCols.fget.cache_clear()
        Sheet.availColnames.fget.cache_clear()
        Sheet.colsByName.fget.cache_clear()

        return cols[0]

    def addColumnAtCursor(self, *cols):
        'Insert all *cols* into columns after cursor.  Return first column.'
        index = 0
        ccol = self.cursorCol
        if ccol and not ccol.keycol:
            try:
                index = self.columns.index(ccol)+1
            except ValueError: # when all columns are hidden, the one column shown is not in self.columns
                index = 0

        self.addColumn(*cols, index=index)
        firstnewcol = [c for c in cols if not c.hidden][0]
        self.cursorVisibleColIndex = self.visibleCols.index(firstnewcol)
        return firstnewcol

    def setColNames(self, rows):
        for c in self.visibleCols:
            c.name = '\n'.join(str(c.getDisplayValue(r)) for r in rows)

    def setKeys(self, cols):
        'Make all *cols* into key columns.'
        vd.addUndo(undoAttrFunc(cols, 'keycol'))
        lastkeycol = 0
        if self.keyCols:
            lastkeycol = max(c.keycol for c in self.keyCols)
        for col in cols:
            if not col.keycol:
                col.keycol = lastkeycol+1
                lastkeycol += 1

    def unsetKeys(self, cols):
        'Make all *cols* non-key columns.'
        vd.addUndo(undoAttrFunc(cols, 'keycol'))
        for col in cols:
            col.keycol = 0

    def rowkey(self, row):
        'Return tuple of the key for *row*.'
        return tuple(c.getTypedValue(row) for c in self.keyCols)

    def rowname(self, row):
        'Return string of the key for *row*.'
        return ','.join(map(str, self.rowkey(row)))

    def checkCursor(self):
        'Keep cursor in bounds of data and screen.'
        # keep cursor within actual available rowset
        if self.nRows == 0 or self.cursorRowIndex <= 0:
            self.cursorRowIndex = 0
        elif self.cursorRowIndex >= self.nRows:
            self.cursorRowIndex = self.nRows-1

        if self.cursorVisibleColIndex <= 0:
            self.cursorVisibleColIndex = 0
        elif self.cursorVisibleColIndex >= len(self.availCols):
            self.cursorVisibleColIndex = len(self.availCols)-1

        if self.topRowIndex < 0:
            self.topRowIndex = 0
        elif self.topRowIndex > self.nRows-1:
            self.topRowIndex = self.nRows-1

        self.adjustColLayout()

        # calculations that rely on nScreenRows, like bottomRowIndex, need to be done after
        # col layout has been adjusted. nScreenRows requires an accurate count of
        # allAggregators, which requires knowing col visibility.
        # check bounds, scroll if necessary
        if self.topRowIndex > self.cursorRowIndex:
            self.topRowIndex = self.cursorRowIndex
        elif self.bottomRowIndex < self.cursorRowIndex:
            self.bottomRowIndex = self.cursorRowIndex

    def adjustColLayout(self):
        '''Move the left visible column to try to keep the cursorCol visible.
        though the cursorCol cannot be visible when screen is totally filled by keycols.
        Run calcColLayout() at least once.'''
        # jumping to a column left of the previously on-screen columns:   put cursorCol as leftmost col
        # jumping to a column right of the previously on-screen columns:  put cursorCol as far right as possible
        if self.leftVisibleColIndex > self.cursorVisibleColIndex:        # e.g. when jumping/moving left
            self.leftVisibleColIndex = self.cursorVisibleColIndex
        elif self.leftVisibleColIndex < self.cursorVisibleColIndex:      # e.g. when jumping/moving right
            #move leftVisibleCol until the cursor column fits fully on screen
            while self.leftVisibleColIndex < self.cursorVisibleColIndex:  #ensures termination even if screen is completely filled by keycols
                self.calcColLayout()
                if not self._visibleColLayout:
                    break

                # If the cursor is outside the visible columns currently laid out (1 window wide).
                # One way to trigger this is with zc, jump to a column never seen yet.
                mincolidx, maxcolidx = min(self._visibleColLayout.keys()), max(self._visibleColLayout.keys())
                if self.cursorVisibleColIndex < mincolidx:
                    # This case is expected never to occur. _visibleColLayout keys are enumerated from 0,
                    # so mincolidx is always 0. and cursorVisibleColIndex is kept >= 0 (by checkCursor).
                    self.leftVisibleColIndex = self.cursorVisibleColIndex
                    break
                elif self.cursorVisibleColIndex > maxcolidx:
                    # some cases:  1) jumping rightward, so cursor has just moved to a column that is offscreen to the right
                    #              2) when keycols fill entire screen
                    self.leftVisibleColIndex += 1
                    continue

                cur_x, cur_w = self._visibleColLayout[self.cursorVisibleColIndex]
                if cur_x+cur_w < self.windowWidth-1:  # current columns fit entirely on screen
                    break
                self.leftVisibleColIndex += 1  # once within the bounds, walk over one column at a time

        if self.leftVisibleColIndex == self.cursorVisibleColIndex:  #will happen after cursor: jumped left, jumped right, or stayed in place
            self.calcColLayout()

    def calcColLayout(self):
        '''Set right-most visible column, based on calculation.
        Assign x coordinates and width to every column that fits on screen, visible or hidden.'''
        minColWidth = dispwidth(self.options.disp_more_left)+dispwidth(self.options.disp_more_right)+2
        sepColWidth = dispwidth(self.options.disp_column_sep)
        winWidth = self.windowWidth
        self._visibleColLayout = {}
        x = 0
        vcolidx = 0
        for vcolidx, col in enumerate(self.availCols):
            width = self.calcSingleColLayout(col, vcolidx, x, minColWidth)
            if width is not None:
                if x < winWidth-1:
                    self._visibleColLayout[vcolidx] = [x, width]
                    x += width+sepColWidth
            if x >= winWidth-1:
                break

        self.rightVisibleColIndex = vcolidx

    def calcSingleColLayout(self, col:Column, vcolidx:int, x:int=0, minColWidth:int=4):
            '''Return the width, for key columns, or for columns that are rightward of
            the leftmost visibleCol, even if they are offscreen or hidden. Return
            None for columns left of cursorVisibleColIndex, if they are not key columns.'''
            # We use a slice of rows that is similar to self.visibleRows but simpler,
            # and larger. The goal is to avoid using nFooterRows. Because nFooterRows
            # cannot in general be calculated properly until after calcColLayout() has
            # determined which columns are visible.
            vrows = self.rows[self.topRowIndex:self.topRowIndex+self.windowHeight]
            if col.width is None and len(vrows) > 0:
                measure_rows = vrows if self.nRows > 1000 else self.rows[:1000]  #1964
                # handle delayed column width-finding
                col.width = max(col.getMaxWidth(measure_rows), minColWidth)
                if vcolidx < self.nVisibleCols-1:  # let last column fill up the max width
                    col.width = min(col.width, self.options.default_width)

            width = col.width if col.width is not None else self.options.default_width

            # when cursor showing a hidden column
            if vcolidx >= self.nVisibleCols and vcolidx == self.cursorVisibleColIndex:
                width = self.options.default_width

            #subtract 1 character of empty space from windowWidth, for the margin to the right of the sheet
            width = min(width, self.windowWidth-x-1)
            width = max(width, 1)
            if col in self.keyCols or vcolidx >= self.leftVisibleColIndex:  # visible columns
                return width


    def drawColHeader(self, scr, y, h, vcolidx):
        'Compose and draw column header for given vcolidx.'
        col = self.availCols[vcolidx]

        # hdrattr highlights whole column header
        # sepattr is for header separators and indicators
        sepcattr = update_attr(colors.color_default, colors.get_color('color_column_sep'), 2)

        hdrcattr = self._colorize(col, None)
        if vcolidx == self.cursorVisibleColIndex:
            hdrcattr = update_attr(hdrcattr, colors.color_current_hdr, 2)

        C = self.options.disp_column_sep
        if (self.keyCols and col is self.keyCols[-1]) or vcolidx == self.nVisibleCols-1:
            C = self.options.disp_keycol_sep

        x, colwidth = self._visibleColLayout[vcolidx]

        # AnameTC
        T = vd.getType(col.type).icon
        if T is None:  # still allow icon to be explicitly non-displayed ''
            T = '?'

        hdrs = col.name.split('\n')
        for i in range(h):
            name = ''
            if colwidth > 2:
                name = ' '  # save room at front for LeftMore or sorted arrow

            if h-i-1 < len(hdrs):
                name += hdrs[::-1][h-i-1]

            if i == h-1:
                hdrcattr = update_attr(hdrcattr, colors.color_bottom_hdr, 5)

            if y+i < self.windowHeight:
                clipdraw(scr, y+i, x, name, hdrcattr, w=colwidth)
            vd.onMouse(scr, x, y+i, colwidth, 1, BUTTON3_RELEASED='rename-col')

            if C and x+colwidth+dispwidth(C) < self.windowWidth and y+i < self.windowHeight:
                scr.addstr(y+i, x+colwidth, C, sepcattr.attr)

        clipdraw(scr, y+h-1, min(x+colwidth, self.windowWidth-1)-dispwidth(T), T, hdrcattr)

        try:
            if vcolidx == self.leftVisibleColIndex and col not in self.keyCols and self.nonKeyVisibleCols.index(col) > 0:
                A = self.options.disp_more_left
                scr.addstr(y, x, A, sepcattr.attr)
        except ValueError:  # from .index
            pass

        try:
            A = ''
            for j, (sortcol, sortdir) in enumerate(self._ordering):
                if isinstance(sortcol, str):
                    sortcol = self.colsByName.get(sortcol)  # self.column will fail if sortcol was renamed
                if col is sortcol:
                    A = self.options.disp_sort_desc[j] if sortdir else self.options.disp_sort_asc[j]
                    scr.addstr(y+h-1, x, A, hdrcattr.attr)
                    break
        except IndexError:
            pass

    def isVisibleIdxKey(self, vcolidx):
        'Return boolean: is given column index a key column?'
        return self.visibleCols[vcolidx] in self.keyCols

    @drawcache_property
    def allAggregators(self):
        'Return dict of aggname -> list of cols with that aggregator.'
        allaggs = collections.defaultdict(list) # aggname -> list of cols with that aggregator
        for vcolidx, (x, colwidth) in sorted(self._visibleColLayout.items()):
            col = self.availCols[vcolidx]
            if not col.hidden:
                for aggr in col.aggregators:
                    allaggs[aggr.name].append(vcolidx)
        return allaggs

    def draw(self, scr):
        'Draw entire screen onto the `scr` curses object.'
        if not self.columns:
            return

        drawparams = {
            'isNull': self.isNullFunc(),

            'topsep': self.options.disp_rowtop_sep,
            'midsep': self.options.disp_rowmid_sep,
            'botsep': self.options.disp_rowbot_sep,
            'endsep': self.options.disp_rowend_sep,
            'keytopsep': self.options.disp_keytop_sep,
            'keymidsep': self.options.disp_keymid_sep,
            'keybotsep': self.options.disp_keybot_sep,
            'endtopsep': self.options.disp_endtop_sep,
            'endmidsep': self.options.disp_endmid_sep,
            'endbotsep': self.options.disp_endbot_sep,

            'colsep': self.options.disp_column_sep,
            'keysep': self.options.disp_keycol_sep,
            'selectednote': self.options.disp_selected_note,
            'disp_truncator': self.options.disp_truncator,
        }

        self._rowLayout = {}  # [rowidx] -> (y, height)
        self.calcColLayout()

        numHeaderRows = self.nHeaderRows
        vcolidx = 0

        headerRow = 0
        for vcolidx, colinfo in sorted(self._visibleColLayout.items()):
            self.drawColHeader(scr, headerRow, numHeaderRows, vcolidx)

        y = headerRow + numHeaderRows

        rows = self.rows[self.topRowIndex:min(self.topRowIndex+self.nScreenRows, self.nRows)]
        vd.callNoExceptions(self.checkCursor)

        for rowidx, row in enumerate(rows):
            if y >= self.windowHeight-1:
                break

            rowcattr = self._colorize(None, row)

            y += self.drawRow(scr, row, self.topRowIndex+rowidx, y, rowcattr, maxheight=self.windowHeight-y-1, **drawparams)

        if vcolidx+1 < self.nVisibleCols:
            scr.addstr(headerRow, self.windowWidth-2, self.options.disp_more_right, colors.color_column_sep.attr)

        # draw bottom-row aggregators  #2209
        rightx, rightw = self._visibleColLayout[self.rightVisibleColIndex]
        rightx += rightw+1

        for aggrname, colidxs in self.allAggregators.items():
            clipdraw(scr, y, 0, ' '*rightx + f' {aggrname:9}', colors.color_aggregator, truncator='+')

            for vcolidx in colidxs:
                x, colwidth = self._visibleColLayout[vcolidx]
                col = self.availCols[vcolidx]

                if not col.hidden:
                    dw = DisplayWrapper('')
                    try:
                        agg = vd.aggregators[aggrname]
                        dw.value = col.aggregateTotal(agg)
                        dw.typedval = wrapply(agg.type or col.type, dw.value)
                        dw.text = col.format(dw.typedval)
                    except Exception as e:
                        dw.note = self.options.disp_note_typeexc
                        dw.notecolor = 'color_warning'
                        vd.exceptionCaught(e, status=False)
                    disps = [('', ' ')] + list(col.display(dw, width=colwidth))
                    clipdraw_chunks(scr, y, x, disps, colors.color_aggregator, w=colwidth)
            y += 1


    def calc_height(self, row, displines=None, isNull=None, maxheight=1):
            'render cell contents for row into displines'
            if displines is None:
                displines = {}  # [vcolidx] -> list of lines in that cell

            for vcolidx, (x, colwidth) in sorted(self._visibleColLayout.items()):
                if x < self.windowWidth:  # only draw inside window
                    vcols = self.availCols
                    if vcolidx >= self.nVisibleCols and vcolidx != self.cursorVisibleColIndex:
                        continue
                    col = vcols[vcolidx]
                    cellval = col.getCell(row)

                    cellval.display = col.display(cellval, colwidth)

                    try:
                        if isNull and isNull(cellval.value):
                            cellval.note = self.options.disp_note_none
                            cellval.notecolor = 'color_note_type'
                    except (TypeError, ValueError):
                        pass

                    if maxheight > 1:
                        lines = _splitcell(self, cellval.display, width=colwidth-2, maxheight=maxheight)
                    else:
                        lines = [cellval.display]
                    displines[vcolidx] = (col, cellval, lines)

            if len(displines) == 0:
                return 0
            return max(len(lines) for _, _, lines in displines.values())

    def drawRow(self, scr, row, rowidx, ybase, rowcattr: ColorAttr, maxheight,
            isNull='',
            topsep='',
            midsep='',
            botsep='',
            endsep='',
            keytopsep='',
            keymidsep='',
            keybotsep='',
            endtopsep='',
            endmidsep='',
            endbotsep='',
            colsep='',
            keysep='',
            selectednote='',
            disp_truncator=''
       ):
            # sepattr is the attr between cell/columns
            sepcattr = update_attr(rowcattr, colors.color_column_sep, 1)

            # apply current row here instead of in a colorizer, because it needs to know dispRowIndex
            if rowidx == self.cursorRowIndex:
                color_current_row = colors.get_color('color_current_row', 2)
                basecellcattr = sepcattr = update_attr(rowcattr, color_current_row)
            else:
                basecellcattr = rowcattr

            # calc_height renders cell contents into displines
            displines = {}  # [vcolidx] -> list of lines in that cell
            if options.disp_multiline_focus:
                height = self.rowHeight if rowidx == self.cursorRowIndex else 1
            else:
                height = min(self.rowHeight, maxheight) or 1  # display even empty rows

            self.calc_height(row, displines, maxheight=height)

            self._rowLayout[rowidx] = (ybase, height)

            if height > 1:
                colseps = [topsep] + [midsep]*(height-2) + [botsep]
                endseps = [endtopsep] + [endmidsep]*(height-2) + [endbotsep]
                keyseps = [keytopsep] + [keymidsep]*(height-2) + [keybotsep]
                color_multiline_bottom = colors.get_color('color_multiline_bottom', 2)
            else:
                colseps = [colsep]
                endseps = [endsep]
                keyseps = [keysep]
                color_multiline_bottom = 0

            for vcolidx, (col, cellval, lines) in displines.items():
                    if vcolidx not in self._visibleColLayout:
                        continue

                    if vcolidx == self.nVisibleCols-1:  # right edge of sheet
                        seps = endseps
                    elif (self.keyCols and col is self.keyCols[-1]): # last keycol
                        seps = keyseps
                    else:
                        seps = colseps

                    x, colwidth = self._visibleColLayout[vcolidx]
                    hoffset = col.hoffset
                    voffset = col.voffset

                    cattr = self._colorize(col, row, cellval)
                    cattr = update_attr(cattr, basecellcattr)
                    bottomcattr = update_attr(cattr, color_multiline_bottom) if height > 1 else cattr

                    note = getattr(cellval, 'note', None)
                    notewidth = 1 if note else 0
                    if note:
                        notecattr = update_attr(cattr, colors.get_color(cellval.notecolor), 10)
                        clipdraw(scr, ybase, x+colwidth-notewidth, note, notecattr)

                    lines = lines[voffset:]

                    if len(lines) > height:
                        lines = lines[:height]
                    elif len(lines) < height:
                        lines.extend([[('', '')]]*(height-len(lines)))

                    for i, chunks in enumerate(lines):
                        y = ybase+i

                        sepchars = seps[i]

                        pre = disp_truncator if hoffset != 0 else disp_column_fill
                        prechunks = []
                        if colwidth > 2:
                            prechunks.append(('', pre))

                        for attr, text in chunks:
                            prechunks.append((attr, text[hoffset:]))

                        clipdraw_chunks(scr, y, x, prechunks, cattr if i < height-1 else bottomcattr, w=colwidth-notewidth)
                        vd.onMouse(scr, x, y, colwidth, 1, BUTTON3_RELEASED='edit-cell')

                        if sepchars and x+colwidth+dispwidth(sepchars) <= self.windowWidth-1:
                            scr.addstr(y, x+colwidth, sepchars, sepcattr.attr)

            for notefunc in vd.rowNoters:
                ch = notefunc(self, row)
                if ch:
                    clipdraw(scr, ybase, 0, ch, colors.color_note_row)
                    break

            return height

    def incremented_colname(self):
        vd.addUndo(setattr, self, 'colname_ctr', self.colname_ctr)
        self.colname_ctr += 1
        return colname_letters(self.colname_ctr)

vd.rowNoters = [
    # f(sheet, row) -> character to be displayed on the left side of row
]

Sheet = TableSheet  # deprecated in 2.0 but still widely used internally


class SequenceSheet(Sheet):
    'Sheets with ``ColumnItem`` columns, and rows that are Python sequences (list, namedtuple, etc).'
    def setCols(self, headerrows):
        self.columns = []
        vd.clearCaches()  #1997
        for i, colnamelines in enumerate(itertools.zip_longest(*headerrows, fillvalue='')):
            colnamelines = ['' if c is None else c for c in colnamelines]
            self.addColumn(ColumnItem(''.join(map(str, colnamelines)), i))

        self._rowtype = namedlist('tsvobj', [(c.name or '_') for c in self.columns])

    def newRow(self):
        return self._rowtype()

    def addRow(self, row, index=None):
        for i in range(len(self.columns), len(row)):  # no-op if already done
            self.addColumn(ColumnItem('', i))
            self._rowtype = namedlist('tsvobj', [(c.name or '_') for c in self.columns])
        if type(row) is not self._rowtype:
            row = self._rowtype(row)
        super().addRow(row, index=index)

    def optlines(self, it, optname):
        'Generate next options.<optname> elements from iterator with exceptions wrapped.'
        for i in range(self.options.getobj(optname, self)):
            try:
                yield next(it)
            except StopIteration:
                break

    def loader(self):
        'Skip first options.skip rows; set columns from next options.header rows.'

        itsource = self.iterload()

        # skip the first options.skip rows
        list(self.optlines(itsource, 'skip'))

        # use the next options.header rows as columns
        self.setCols(list(self.optlines(itsource, 'header')))

        self.rows = []
        # add the rest of the rows
        max_rows = self.options.max_rows
        for i, r in enumerate(vd.Progress(itsource, gerund='loading', total=0)):
            if self.precious and i >= max_rows:
                break
            self.addRow(r)


@VisiData.property
@drawcache
def _evalcontexts(vd):
    return {}

## VisiData sheet manipulation

@VisiData.api
def replace(vd, vs):
    'Replace top sheet with the given sheet `vs`.'
    vd.sheets.pop(0)
    return vd.push(vs)


@VisiData.api
def remove(vd, vs):
    'Remove *vs* from sheets stack, without asking for confirmation.'
    if vs in vd.sheets:
        vd.sheets.remove(vs)
        if vs in vd.allSheets:
            vd.allSheets.remove(vs)
            vd.allSheets.append(vs)
    else:
        vd.fail('sheet not on stack')


@VisiData.api
def push(vd, vs, pane=0, load=True):
    'Push Sheet *vs* onto ``vd.sheets`` stack for *pane* (0 for active pane, -1 for inactive pane).  Remove from other position if already on sheets stack.'
    if not isinstance(vs, BaseSheet):
        return  # return instead of raise, some commands need this

    if vs in vd.sheets:
        vd.sheets.remove(vs)

    vs.vd = vd
    if pane == -1:
        vs.pane = 2 if vd.activePane == 1 else 1
    elif pane == 0:
        if not vd.sheetstack(1): vd.activePane=vs.pane=1
        elif not vd.sheetstack(2) and vd.options.disp_splitwin_pct != 0: vd.activePane=vs.pane=2
        else: vs.pane = vd.activePane
    else:
        vs.pane = pane

    vd.sheets.insert(0, vs)

    if vs.precious and vs not in vd.allSheets:
        vd.allSheets.append(vs)

    if load:
        vs.ensureLoaded()
    if vd.activeCommand:
        vs.longname = vd.activeCommand.longname


@VisiData.api
def quit(vd, *sheets):
    'Remove *sheets* from sheets stack, asking for confirmation if needed.'

    for vs in sheets:
        vs.confirmQuit('quit')
        vs.pane = 0
        vd.remove(vs)
    if vd.activeCommand:
        vd.activeSheet.longname = vd.activeCommand.longname


@BaseSheet.api
def confirmQuit(vs, verb='quit'):
    if vs.options.quitguard and vs.precious and vs.hasBeenModified and not vd._nextCommands:
        vd.draw_all()
        vd.confirm(f'{verb} modified sheet "{vs.name}"? ')
    elif vs.options.getonly('quitguard', vs, False) and not vd._nextCommands:  # if this sheet is specifically guarded
        vd.draw_all()
        vd.confirm(f'{verb} guarded sheet "{vs.name}"? ')


@BaseSheet.api
def preloadHook(sheet):
    'Override to setup for reload().'
    sheet.confirmQuit('reload')

    sheet.hasBeenModified = False


@VisiData.api
def newSheet(vd, name, ncols, **kwargs):
    cols = [SettableColumn(width=vd.options.default_width, name=f'{colname_letters(i+1)}') for i in range(ncols)]
    vs = Sheet(name, columns=cols, **kwargs)
    vs.colname_ctr = ncols
    return vs

@BaseSheet.api
def quitAndReleaseMemory(vs):
    'Release largest memory consumer refs on *vs* to free up memory.'
    if isinstance(vs.source, visidata.Path):
        vs.source.lines.clear() # clear cache of read lines

    if vs.precious: # only precious sheets have meaningful data
        vs.confirmQuit('quit')
        vs.rows.clear()
        vs.rows = UNLOADED
        vd.remove(vs)
        vd.allSheets.remove(vs)


@BaseSheet.api
def splitPane(sheet, pct=None):
    if vd.activeStack[1:]:
        undersheet = vd.activeStack[1]
        pane = 1 if undersheet.pane == 2 else 2
        vd.push(undersheet, pane=pane)
        vd.activePane = pane

    vd.options.disp_splitwin_pct = pct


@Sheet.api
def async_deepcopy(sheet, rowlist):
    @asyncthread
    def _async_deepcopy(newlist, oldlist):
        for r in vd.Progress(oldlist, 'copying'):
            newlist.append(deepcopy(r))

    ret = []
    _async_deepcopy(ret, rowlist)
    return ret

@Sheet.api
def reload_or_replace(sheet):
    sheet.preloadHook()
    if isinstance(sheet.source, visidata.Path) and \
       sheet.source.is_url() and sheet.source.scheme != 'file':  #2825
        #retrieve data again, because the earlier data saved in sheet.source may be outdated
        vs = vd.openSource(visidata.Path(sheet.source.given))
        if type(vs) != type(sheet):  #new data may have a different filetype
            vd.push(vs)
            vd.remove(sheet)
            #user needs feedback that sheet changed, since the new sheet has a different shortcut
            vd.status('replaced sheet due to changed filetype')
            return
        sheet.source = vs.source
    sheet.reload()


BaseSheet.init('pane', lambda: 1)


BaseSheet.addCommand('^R', 'reload-sheet', 'reload_or_replace()', 'Reload current sheet')
Sheet.addCommand('', 'show-cursor', 'status(statusLine)', 'show cursor position and bounds of current sheet on status line')

Sheet.addCommand('!', 'key-col', 'exec_longname("key-col-off") if cursorCol.keycol else exec_longname("key-col-on")', 'toggle current column as a key column', replay=False)
Sheet.addCommand('', 'key-col-on', 'setKeys([cursorCol])', 'set current column as a key column')
Sheet.addCommand('z!', 'key-col-off', 'unsetKeys([cursorCol])', 'unset current column as a key column')

Sheet.addCommand('e', 'edit-cell', 'cursorCol.setValues([cursorRow], editCell(cursorVisibleColIndex)) if not (cursorRow is None) else fail("no rows to edit")', 'edit contents of current cell')
Sheet.addCommand('ge', 'setcol-input', 'cursorCol.setValuesTyped(selectedRows, input("set selected to: ", value=cursorDisplay))', 'set contents of current column for selected rows to same input')

Sheet.addCommand('"', 'dup-selected', 'vs=copy(sheet); vs.name += "_selectedref"; vs.reload=lambda vs=vs,rows=selectedRows: setattr(vs, "rows", list(rows)); vd.push(vs)', 'open a duplicate sheet with only the selected rows')
Sheet.addCommand('g"', 'dup-rows', 'vs=copy(sheet); vs.name+="_copy"; vs.rows=list(rows); status("copied "+vs.name); vs.select(selectedRows); vd.push(vs)', 'open a duplicate sheet with all rows')
Sheet.addCommand('z"', 'dup-selected-deep', 'vs = deepcopy(sheet); vs.name += "_selecteddeepcopy"; vs.rows = vs.async_deepcopy(selectedRows); vd.push(vs); status("pushed sheet with async deepcopy of selected rows")', 'open duplicate sheet with deepcopy of selected rows')
Sheet.addCommand('gz"', 'dup-rows-deep', 'vs = deepcopy(sheet); vs.name += "_deepcopy"; vs.rows = vs.async_deepcopy(rows); vd.push(vs); status("pushed sheet with async deepcopy of all rows")', 'open duplicate sheet with deepcopy of all rows')

Sheet.addCommand('z~', 'type-any', 'cursorCol.type = anytype', 'set type of current column to anytype')
Sheet.addCommand('~', 'type-string', 'cursorCol.type = str', 'set type of current column to str')
Sheet.addCommand('#', 'type-int', 'cursorCol.type = int', 'set type of current column to int')
Sheet.addCommand('z#', 'type-len', 'cursorCol.type = vlen', 'set type of current column to len')
Sheet.addCommand('%', 'type-float', 'cursorCol.type = float', 'set type of current column to float')
Sheet.addCommand('', 'type-floatlocale', 'cursorCol.type = floatlocale', 'set type of current column to float using system locale set in LC_NUMERIC')

BaseSheet.addCommand('q', 'quit-sheet',  'vd.quit(sheet)', 'quit current sheet')
BaseSheet.addCommand('Q', 'quit-sheet-free',  'quitAndReleaseMemory()', 'discard current sheet and free memory')
globalCommand('gq', 'quit-all', 'vd.quit(*vd.sheets)', 'quit all sheets (clean exit)')

BaseSheet.addCommand('Z', 'splitwin-half', 'splitPane(vd.options.disp_splitwin_pct or 50)', 'ensure split pane is set and push under sheet onto other pane')
BaseSheet.addCommand('gZ', 'splitwin-close', 'vd.options.disp_splitwin_pct = 0\nfor vs in vd.activeStack: vs.pane = 1', 'close split screen')
BaseSheet.addCommand('^I', 'splitwin-swap', 'vd.activePane = 1 if sheet.pane == 2 else 2', 'jump to inactive pane')
BaseSheet.addCommand('g^I', 'splitwin-swap-pane', 'vd.options.disp_splitwin_pct=-vd.options.disp_splitwin_pct', 'swap panes onscreen')
BaseSheet.addCommand('zZ', 'splitwin-input', 'vd.options.disp_splitwin_pct = input("% height for split window: ", value=vd.options.disp_splitwin_pct)', 'set split pane to specific size')

BaseSheet.addCommand('^L', 'redraw', 'sheet.refresh(); vd.redraw(); vd.draw_all()', 'Refresh screen')
BaseSheet.addCommand(None, 'guard-sheet', 'options.set("quitguard", True, sheet); status("guarded")', 'Set quitguard on current sheet to confirm before quit')
BaseSheet.addCommand(None, 'guard-sheet-off', 'options.set("quitguard", False, sheet); status("unguarded")', 'Unset quitguard on current sheet to not confirm before quit')
BaseSheet.addCommand(None, 'open-source', 'vd.replace(source)', 'jump to the source of this sheet')

BaseSheet.bindkey('KEY_RESIZE', 'redraw')

BaseSheet.addCommand('A', 'open-new', 'vd.push(vd.newSheet("unnamed", 1))', 'Open new empty sheet')

BaseSheet.addCommand('`', 'open-source', 'vd.push(source)', 'open source sheet')
BaseSheet.addCommand(None, 'rename-sheet', 'sheet.name = input("rename sheet to: ", value=cleanName(sheet.name))', 'Rename current sheet')

Sheet.addCommand('', 'addcol-source', 'source.addColumn(copy(cursorCol)) if isinstance (source, BaseSheet) else error("source must be sheet")', 'add copy of current column to source sheet')  #988  frosencrantz


@Column.api
def formatter_enum(col, fmtdict):
    return lambda val, fmtdict=fmtdict,*args,**kwargs: fmtdict.__getitem__(val)

Sheet.addCommand('', 'setcol-formatter', 'cursorCol.formatter=input("set formatter to: ", value=cursorCol.formatter or "generic")', 'set formatter for current column (generic, json, python)')
Sheet.addCommand('', 'setcol-format-enum', 'cursorCol.fmtstr=input("format replacements (k=v): ", value=f"{cursorDisplay}=", i=len(cursorDisplay)+1); cursorCol.formatter="enum"', 'add secondary type translator to current column from input enum (space-separated)')


vd.addGlobals(
    RowColorizer=RowColorizer,
    CellColorizer=CellColorizer,
    ColumnColorizer=ColumnColorizer,
    RecursiveExprException=RecursiveExprException,
    LazyComputeRow=LazyComputeRow,
    Sheet=Sheet,
    TableSheet=TableSheet,
    SequenceSheet=SequenceSheet)

vd.addMenuItems('''
    File > New > open-new
    File > Rename > rename-sheet
    File > Guard > on > guard-sheet
    File > Guard > off > guard-sheet-off
    File > Duplicate > selected rows by ref > dup-selected
    File > Duplicate > all rows by ref > dup-rows
    File > Duplicate > selected rows deep > dup-selected-deep
    File > Duplicate > all rows deep > dup-rows-deep
    File > Reload > rows and columns > reload-sheet
    File > Quit > top sheet > quit-sheet
    File > Quit > all sheets > quit-all
    Edit > Modify > current cell > input > edit-cell
    Edit > Modify > selected cells > from input > setcol-input
    View > Sheets > stack > sheets-stack
    View > Sheets > all > sheets-all
    View > Other sheet > source sheet > open-source
    View > Split pane > in half > splitwin-half
    View > Split pane > in percent > splitwin-input
    View > Split pane > unsplit > splitwin-close
    View > Split pane > swap panes > splitwin-swap-pane
    View > Split pane > goto other pane > splitwin-swap
    View > Refresh screen > redraw
    Column > Type as > anytype > type-any
    Column > Type as > string > type-string
    Column > Type as > integer > type-int
    Column > Type as > float > type-float
    Column > Type as > locale float > type-floatlocale
    Column > Type as > length > type-len
    Column > Key > toggle current column > key-col
    Column > Key > unkey current column > key-col-off
''')



from visidata import VisiData, vd, Sheet

vd.si_prefixes='p n u m . kK M G T P Q'.split()


@VisiData.api
def SIFormatter(vd, fmtstr, val):
    level = 4
    if val != 0:
        while abs(val) > 1000:
            val /= 1000
            level += 1
        while abs(val) < 0.001:
            val *= 1000
            level -= 1

    return vd.numericFormatter(fmtstr, val) + (vd.si_prefixes[level][0] if level != 4 else '')


@vd.numericType('‱', formatter=vd.SIFormatter)
def floatsi(*args):
    if not args:
        return 0.0
    if not isinstance(args[0], str):
        return float(args[0])

    s=args[0].strip()
    for i, p in enumerate(vd.si_prefixes):
        if s[-1] in p:
            return float(s[:-1]) * (1000 ** (i-4))

    return float(s)


Sheet.addCommand('z%', 'type-floatsi', 'cursorCol.type = floatsi', 'set type of current column to SI float')

vd.addMenuItems('''
    Column > Type as > SI float > type-floatsi
''')



'''
# Using VisiData with Google Sheets and Google Drive

## Setup and Authentication

Add to .visidatarc:

    import visidata.experimental.google

When VisiData attempts to use the Google API, it uses the "web authentication flow", which causes a web page to open asking for permissions to read and/or write your Google Sheets.
After granting permissions, VisiData caches the auth token in the .visidata directory.  Remove `.visidata/google-*.pickle` to unauthenticate.

## Load a Google Sheet into VisiData

Use VisiData to open the URL or spreadsheet id, with filetype `g` (or `gsheets`):

    vd https://docs.google.com/spreadsheets/d/1WV0JI_SsGfmoocXWJILK2nhfcxU1H7roqL1HE7zBdsY/ -f g

VisiData assumes the first row is the header row with column names.

## Save one or more sheets in VisiData as a Google Sheet

Save to `<sheetname>.g` using either `Ctrl+S` (current sheet only) or `g Ctrl+S` (all sheets on the sheet stack).
<sheetname> will be the visible name of the Spreadsheet in Google Drive; each sheet tab within the Spreadsheet will be named according to the sheet name within VisiData.

## List files in Google Drive

Use the `gdrive` filetype (the path doesn't matter):

    vd . -f gdrive

- Files can be marked for deletion with `d` and execute those deletions with `z Ctrl+S` (same as on the DirSheet for the local filesystem).
- Images can be viewed with `Enter` (in browser).
'''

import visidata.experimental.gdrive
import visidata.experimental.gsheets



import os
import sys
import signal
import subprocess
import tempfile
import curses

import visidata

visidata.vd.tstp_signal = None

class SuspendCurses:
    'Context manager to leave windowed mode on enter and restore it on exit.'
    def __enter__(self):
        if visidata.vd.scrFull:
            curses.endwin()
        if visidata.vd.tstp_signal:
            signal.signal(signal.SIGTSTP, visidata.vd.tstp_signal)

    def __exit__(self, exc_type, exc_val, tb):
        if visidata.vd.scrFull:
            curses.reset_prog_mode()
            visidata.vd.scrFull.refresh()
            curses.doupdate()


@visidata.VisiData.api
def launchEditor(vd, *args):
    'Launch $EDITOR with *args* as arguments.'
    editor = os.environ.get('EDITOR') or vd.fail('$EDITOR not set')
    args = editor.split() + list(args)
    with SuspendCurses():
        return subprocess.call(args)


@visidata.VisiData.api
def launchBrowser(vd, *args):
    'Launch $BROWSER with *args* as arguments.'
    browser = os.environ.get('BROWSER') or vd.fail('no $BROWSER for %s' % args[0])
    vd.status('opening ' + args[0])
    args = [browser] + list(args)
    subprocess.call(args)


@visidata.VisiData.api
def launchExternalEditor(vd, v, linenum=0):
    'Launch $EDITOR to edit string *v* starting on line *linenum*.'
    import tempfile
    with tempfile.NamedTemporaryFile() as temp:
        temp.close()  #2118 must close before re-opening on windows
        with open(temp.name, 'w') as fp:
            fp.write(v)
        return vd.launchExternalEditorPath(visidata.Path(temp.name), linenum)


@visidata.VisiData.api
def launchExternalEditorPath(vd, path, linenum=0):
        'Launch $EDITOR to edit *path* starting on line *linenum*.'
        if linenum:
            visidata.vd.launchEditor(path, '+%s' % linenum)
        else:
            visidata.vd.launchEditor(path)

        with open(path, 'r') as fp:
            try:
                return fp.read().rstrip('\n')  # trim inevitable trailing newlines
            except Exception as e:
                visidata.vd.exceptionCaught(e)
                return ''


@visidata.VisiData.api
def suspend(vd):
    import signal
    with SuspendCurses():
        os.kill(os.getpid(), signal.SIGSTOP)


def _breakpoint(*args, **kwargs):
    import pdb
    class VisiDataPdb(pdb.Pdb):
        def precmd(self, line):
            r = super().precmd(line)
            if not r:
                SuspendCurses.__exit__(None, None, None, None)
            return r

        def postcmd(self, stop, line):
            if stop:
                SuspendCurses.__enter__(None)
            return super().postcmd(stop, line)

    SuspendCurses.__enter__(None)
    VisiDataPdb(nosigint=True).set_trace()


sys.breakpointhook = _breakpoint


visidata.BaseSheet.addCommand('^Z', 'suspend', 'suspend()', 'suspend VisiData process')
visidata.BaseSheet.addCommand('', 'breakpoint', 'breakpoint()', 'drop into pdb REPL')

visidata.vd.addGlobals(SuspendCurses=SuspendCurses)



import collections
import re
from copy import copy

from visidata import Sheet, SubColumnItem, ColumnItem, Column, Progress
from visidata import asyncthread, vd

melt_var_colname = 'Variable' # column name to use for the melted variable name
melt_value_colname = 'Value'  # column name to use for the melted value
melt_null = False   # whether to include rows for null values during melt


# rowdef: {0:sourceRow, 1:Category1, ..., N:CategoryN, ColumnName:Column, ...}
class MeltedSheet(Sheet):
    "Perform 'melt', the inverse of 'pivot', on input `source` sheet.  Pass `regex` to parse column names into additional columns"

    rowtype = 'melted values'

    def getValueCols(self) -> dict:
        '''Return dict of ('Category1', 'Category2') -> list of tuple('ColumnName', Column)'''
        colsToMelt = [copy(c) for c in self.source.nonKeyVisibleCols]

        # break down Category1_Category2_ColumnName as per regex
        valcols = collections.OrderedDict()
        for c in colsToMelt:
            c.aggregators = [vd.aggregators['max']]
            m = re.match(self.regex, c.name)
            if m:
                if len(m.groups()) == 1:
                    varvals = m.groups()
                    valcolname = melt_value_colname
                else:
                    *varvals, valcolname = m.groups()
                cats = tuple(varvals)
                if cats not in valcols:
                    valcols[cats] = []
                valcols[cats].append((valcolname, c))
                ncats = len(varvals)
            else:
                vd.status(f'"{c.name}" column does not match regex, skipping')
                ncats = 0

        return valcols

    def resetCols(self):
        self.columns = []
        sheet = self.source
        for c in sheet.keyCols:
            self.addColumn(SubColumnItem(0, c))
        self.setKeys(self.columns)

        valcols = self.getValueCols()
        othercols = set()
        ncats = 0
        for colnames, cols in valcols.items():
            ncats = max(ncats, len(colnames))
            for cname, _ in cols:
                othercols.add(cname)

        if ncats == 1:
            self.addColumn(ColumnItem(melt_var_colname, 1))
        else:
            for i in range(ncats):
                self.addColumn(ColumnItem(f'{melt_var_colname}{i+1}', i+1))

        for cname in othercols:
            self.addColumn(Column(cname,
                getter=lambda col,row,cname=cname: row[cname].getValue(row[0]),
                setter=lambda col,row,val,cname=cname: row[cname].setValues([row[0]], val),
                aggregators=[vd.aggregators['max']]))

    def iterload(self):
        isNull = self.isNullFunc()

        valcols = self.getValueCols()
        for r in Progress(self.source.rows, 'melting'):
            for colnames, cols in valcols.items():
                meltedrow = {}
                for varval, c in cols:
                    try:
                        if melt_null or not isNull(c.getValue(r)):
                            meltedrow[varval] = c
                    except Exception as e:
                        pass

                if meltedrow:  # remove rows with no content (all nulls)
                    meltedrow[0] = r
                    for i, colname in enumerate(colnames):
                        meltedrow[i+1] = colname

                    yield meltedrow


@Sheet.api
def openMelt(sheet, regex='(.*)'):
    return MeltedSheet(sheet.name, 'melted', source=sheet, regex=regex)


Sheet.addCommand('M', 'melt', 'vd.push(openMelt())', 'open Melted Sheet (unpivot), with key columns retained and all non-key columns reduced to Variable-Value rows')

Sheet.addCommand('gM', 'melt-regex', 'vd.push(openMelt(vd.inputRegex("regex to split colname: ", value="(.*)_(.*)", type="regex-capture")))', 'open Melted Sheet (unpivot), with key columns retained and regex capture groups determining how the non-key columns will be reduced to Variable-Value rows')

vd.addMenuItems('''
    Data > Melt > nonkey columns > melt
    Data > Melt > nonkey columns by regex > melt-regex
''')
vd.addGlobals(MeltedSheet=MeltedSheet)



import math
import random
import os.path
from functools import singledispatch

from visidata import vd, Sheet, asyncthread, Progress, Column, VisiData, deduceType, anytype, getitemdef, ColumnsSheet


@Sheet.api
def getSampleRows(sheet):
    'Return list of sample rows, including the cursor row as the first row.'

    # do not include cursorRow in sample
    ret = sheet.rows[:sheet.cursorRowIndex] + sheet.rows[sheet.cursorRowIndex+1:]

    n = sheet.options.default_sample_size
    if n != 0 and n < sheet.nRows:
        vd.aside(f'sampling {n} rows')
        ret = random.sample(ret, n)

    return [sheet.cursorRow] + ret


@Sheet.api
def expandCols(sheet, cols, rows=None, depth=0):
    'expand all visible columns of containers to the given depth (0=fully)'
    ret = []
    if not rows:
        rows = sheet.getSampleRows()

    for col in cols:
        newcols = col.expand(rows)
        if depth != 1:  # countdown not yet complete, or negative (indefinite)
            ret.extend(sheet.expandCols(newcols, rows, depth-1))
    return ret

@singledispatch
def _createExpandedColumns(sampleValue, col, rows):
    '''By default, a column is not expandable. Supported container types for
    sampleValue trigger alternate, type-specific expansions.'''
    return []

@_createExpandedColumns.register(dict)
def _(sampleValue, col, vals):
    '''Build a set of columns to add, using the first occurrence of each key to
    determine column type'''
    newcols = {}

    for val in Progress(vals, 'expanding'):
        if not isinstance(val, dict):  # allow mixed-use columns
            continue
        colsToAdd = set(val).difference(newcols)
        colsToAdd and newcols.update({
            k: deduceType(v)
            for k, v in val.items()
            if k in colsToAdd
        })

    return [
        ExpandedColumn(col.sheet.options.fmt_expand_dict % (col.name, k), type=v, origCol=col, expr=k)
            for k, v in newcols.items()
    ]

def _createExpandedColumnsNamedTuple(col, val):
    return [
        ExpandedColumn(col.sheet.options.fmt_expand_dict % (col.name, k), type=colType, origCol=col, expr=i)
            for i, (k, colType) in enumerate(zip(val._fields, (deduceType(v) for v in val)))
    ]

@_createExpandedColumns.register(list)
@_createExpandedColumns.register(tuple)
def _(sampleValue, col, vals):
    '''Use the longest sequence to determine the number of columns we need to
    create, and their presumed types.  Ignore strings and exceptions. '''
    def lenNoExceptions(v):
        try:
            if isinstance(v, str):
                return 0
            return len(v)
        except Exception as e:
            return 0

    if hasattr(sampleValue, '_fields'):  # looks like a namedtuple
        return _createExpandedColumnsNamedTuple(col, vals[0])

    longestSeq = max(vals, key=lenNoExceptions)
    colTypes = [deduceType(v) for v in longestSeq]
    return [
        ExpandedColumn(col.sheet.options.fmt_expand_list % (col.name, k), type=colType, origCol=col, expr=k)
            for k, colType in enumerate(colTypes)
    ]


@Column.api
def expand(col, rows):
    isNull = col.sheet.isNullFunc()
    nonNulls = [
        col.getTypedValue(row)
        for row in rows
        if not isNull(col.getValue(row))
    ]

    if not nonNulls:
        return []

    # The type of the first non-null value for col determines if and how the
    # column can be expanded.
    expandedCols = _createExpandedColumns(nonNulls[0], col, nonNulls)

    idx = col.sheet.columns.index(col)

    for i, c in enumerate(expandedCols):
        col.sheet.addColumn(c, index=idx+i+1)
    if expandedCols:
        col.hide()
    return expandedCols


@VisiData.api
class ExpandedColumn(Column):
    def calcValue(self, row):
        return getitemdef(self.origCol.getValue(row), self.expr)

    def setValue(self, row, value, setModified=True):
        self.origCol.getValue(row)[self.expr] = value
        if setModified:
            self.origCol.sheet.setModified()


@Sheet.api
@asyncthread
def contract_cols(sheet, cols, depth=1):  # depth == 0 means contract all the way
    'Remove any columns in cols with .origCol, and also remove others in sheet.columns which share those .origCol.  The inverse of expand.'
    vd.addUndo(setattr, sheet, 'columns', sheet.columns)
    for i in range(depth or 10000):
        colsToClose = [c for c in cols if getattr(c, "origCol", None)]

        if not colsToClose:
            break

        origCols = set(c.origCol for c in colsToClose)
        for col in origCols:
            col.width = sheet.options.default_width

        sheet.columns = [col for col in sheet.columns if getattr(col, 'origCol', None) not in origCols]


@Sheet.api
@asyncthread
def expand_cols_deep(sheet, cols, rows=None, depth=0):  # depth == 0 means drill all the way
    return sheet.expandCols(cols, rows=rows, depth=depth)


@ColumnsSheet.api
def contract_source_cols(sheet, cols):
    prefix = os.path.commonprefix([c.name for c in cols])
    ret = ColumnGroup(prefix or 'group', prefix=prefix, sourceCols=cols)
    for c in cols:
        c.origCol = ret
    for vs in sheet.source:
        vd.addUndo(setattr, vs, 'columns', vs.columns)
        vs.columns[:] = [c for c in vs.columns if c not in cols]
    return ret


class ColumnGroup(Column):
    def calcValue(self, row):
        return {c.name[len(self.prefix):]:c.getValue(row) for c in self.sourceCols}

    def expand(self, rows):
        idx = self.sheet.columns.index(self)

        for i, c in enumerate(self.sourceCols):
            self.sheet.addColumn(c, index=idx+i+1)

        self.hide()

        return self.sourceCols


Sheet.addCommand('(', 'expand-col', 'expand_cols_deep([cursorCol], depth=1)', 'expand current column of containers one level')
Sheet.addCommand('g(', 'expand-cols', 'expand_cols_deep(visibleCols, depth=1)', 'expand all visible columns of containers one level')
Sheet.addCommand('z(', 'expand-col-depth', 'expand_cols_deep([cursorCol], depth=int(input("expand depth=", value=0)))', 'expand current column of containers to given depth (0=fully)')
Sheet.addCommand('gz(', 'expand-cols-depth', 'expand_cols_deep(visibleCols, depth=int(input("expand depth=", value=0)))', 'expand all visible columns of containers to given depth (0=fully)')

Sheet.addCommand(')', 'contract-col', 'contract_cols([cursorCol])', 'remove current column and siblings from sheet columns and unhide parent (1 level)')
Sheet.addCommand('g)', 'contract-cols', 'contract_cols(visibleCols)', 'remove all child columns and unhide toplevel parents (1 level)')
Sheet.addCommand('z)', 'contract-col-depth', 'contract_cols([cursorCol], depth=int(input("contract depth=", value=0)))', 'remove current column and siblings from sheet columns and unhide parent, prompting for depth')
Sheet.addCommand('gz)', 'contract-cols-depth', 'contract_cols(visibleCols, depth=int(input("contract depth=", value=0)))', 'remove all child columns and unhide toplevel parents, prompting for depth')

ColumnsSheet.addCommand(')', 'contract-source-cols', 'source[0].addColumn(contract_source_cols(someSelectedRows), index=cursorRowIndex)', 'contract selected columns into column group')  #1702


vd.addMenuItems('''
    Column > Expand > one level > expand-col
    Column > Expand > to depth N > expand-col-depth
    Column > Expand > all columns one level > expand-cols
    Column > Expand > all columns to depth > expand-cols-depth
    Column > Contract > one level > contract-col
    Column > Contract > N levels > contract-col-depth
    Column > Contract > all columns one level > contract-cols
    Column > Contract > all columns N levels > contract-cols-depth
    Column > Contract > selected columns on source sheet > contract-source-cols
''')



from copy import copy, deepcopy
import shutil
import subprocess
import io
import sys
import tempfile
import functools
import os
import itertools
import platform

from visidata import VisiData, vd, asyncthread, SettableColumn
from visidata import Sheet, Path, Column

if (
    # Windows
    sys.platform == 'win32'
    # WSL 2
    or "microsoft-standard-WSL2" in platform.uname().release
    # WSL 1
    or sys.platform == 'linux' and platform.uname().release.endswith("-Microsoft")
):
    syscopy_cmd_default = 'clip.exe'
    syspaste_cmd_default = 'powershell.exe -noprofile -command Get-Clipboard'
elif sys.platform == 'darwin':
    syscopy_cmd_default = 'pbcopy w'
    syspaste_cmd_default = 'pbpaste'
else:
    if 'WAYLAND_DISPLAY' in os.environ:
        syscopy_cmd_default = 'wl-copy'
        syspaste_cmd_default = 'wl-paste'
    else:
        syscopy_cmd_default = 'xclip -selection clipboard -filter'  # xsel --clipboard --input
        syspaste_cmd_default = 'xclip -selection clipboard -o'  # xsel --clipboard

vd.option('clipboard_copy_cmd', syscopy_cmd_default, 'command to copy stdin to system clipboard', sheettype=None)
vd.option('clipboard_paste_cmd', syspaste_cmd_default, 'command to send contents of system clipboard to stdout', sheettype=None)


@Sheet.api
def copyRows(sheet, rows):
    vd.memory.cliprows = rows
    vd.memory.clipcols = list(sheet.visibleCols)
    if not rows:
        vd.warning('no %s selected; clipboard emptied' % sheet.rowtype)
    else:
        vd.status('copied %d %s to clipboard' % (len(rows), sheet.rowtype))

@Sheet.api
def copyCells(sheet, col, rows):
    vd.memory.clipcells = [col.getTypedValue(r) for r in rows]
    if not rows:
        vd.warning('no %s selected; clipboard emptied' % sheet.rowtype)
        return
    vd.status('copied %d %s.%s to clipboard' % (len(rows), sheet.rowtype, col.name))


@Sheet.api
def syscopyValue(sheet, val):
    # pipe val to stdin of clipboard command

    p = subprocess.run(
        sheet.options.clipboard_copy_cmd.split(),
        input=val,
        encoding='utf-8',
        stdout=subprocess.DEVNULL)

    vd.status('copied value to system clipboard')

@Sheet.api
def setColClipboard(sheet):
    if not vd.memory.clipcells:
        vd.warning("nothing to paste from clipcells")
        return
    sheet.cursorCol.setValuesTyped(sheet.onlySelectedRows, *vd.memory.clipcells)

@Sheet.api
def syscopyCells(sheet, cols, rows, filetype=None):
    filetype = filetype or vd.input("copy %d %s as filetype: " % (len(rows), sheet.rowtype), value=sheet.options.save_filetype or 'tsv')
    sheet.syscopyCells_async(cols, rows, filetype)


@Sheet.api
@asyncthread
def syscopyCells_async(sheet, cols, rows, filetype):
    vs = copy(sheet)
    vs.rows = rows or vd.fail('no %s selected' % sheet.rowtype)
    vs.columns = cols

    vd.status(f'copying {vs.nRows} {vs.rowtype} to system clipboard as {filetype}')

    with io.StringIO() as buf:
        with tempfile.NamedTemporaryFile() as temp:
            temp.close()  #2118

            vd.sync(vd.saveSheets(Path(f'{temp.name}.{filetype}', fptext=buf), vs, confirm_overwrite=False))
            subprocess.run(
                sheet.options.clipboard_copy_cmd.split(),
                input=buf.getvalue(),
                encoding='utf-8',
                stdout=subprocess.DEVNULL)


@VisiData.api
def sysclipValue(vd):
    cmd = vd.options.clipboard_paste_cmd
    return subprocess.check_output(vd.options.clipboard_paste_cmd.split()).decode('utf-8')


@VisiData.api
@asyncthread
def pasteFromClipboard(vd, cols, rows):
    text = vd.getLastArgs() or vd.sysclipValue().strip() or vd.fail('nothing to paste from system clipboard')

    vd.addUndoSetValues(cols, rows)
    lines = text.split('\n')
    if not lines:
        vd.warning('nothing to paste from system clipboard')
        return

    vs = cols[0].sheet
    newrows = [vs.newRow() for i in range(len(lines)-len(rows))]
    if newrows:
        rows.extend(newrows)
        vs.addRows(newrows)

    for line, r in zip(lines, rows):
        for v, c in zip(line.split('\t'), cols):
            c.setValue(r, v)


@Sheet.api
def delete_row(sheet, rowidx):
    if not sheet.rows:
        vd.fail("no row to delete")
    if not sheet.defer:
        oldrow = sheet.rows.pop(rowidx)
        vd.addUndo(sheet.rows.insert, rowidx, oldrow)
        # clear the deleted row from selected rows
        if sheet.isSelected(oldrow):
            sheet.addUndoSelection()
            sheet.unselectRow(oldrow)
    else:
        oldrow = sheet.rows[rowidx]
        sheet.rowDeleted(oldrow)

    sheet.setModified()
    return oldrow


@Sheet.api
@asyncthread
def paste_after(sheet, rowidx):
    'Paste rows from *vd.cliprows* at *rowidx*.'
    if not vd.memory.cliprows:  #1793
        vd.warning('nothing to paste from cliprows')
        return

    for col in vd.memory.clipcols[sheet.nVisibleCols:]:
        newcol = SettableColumn()
        newcol.__setstate__(col.__getstate__())
        sheet.addColumn(newcol)

    addedRows = []

    for extrow in vd.memory.cliprows:
        if isinstance(extrow, Column):
            newrow = copy(extrow)
        else:
            newrow = sheet.newRow()
            for col, extcol in zip(sheet.visibleCols, vd.memory.clipcols):
                col.setValue(newrow, extcol.getTypedValue(extrow))

        addedRows.append(newrow)

    sheet.addRows(addedRows, index=rowidx)


Sheet.addCommand('y', 'copy-row', 'copyRows([cursorRow])', 'yank (copy) current row to clipboard')

Sheet.addCommand('p', 'paste-after', 'paste_after(cursorRowIndex)', 'paste clipboard rows after current row')
Sheet.addCommand('P', 'paste-before', 'paste_after(cursorRowIndex-1)', 'paste clipboard rows before current row')

Sheet.addCommand('gy', 'copy-selected', 'copyRows(onlySelectedRows)', 'yank (copy) selected rows to clipboard')

Sheet.addCommand('zy', 'copy-cell', 'copyCells(cursorCol, [cursorRow]); vd.memoValue("clipval", cursorTypedValue, cursorDisplay)', 'yank (copy) current cell to clipboard')
Sheet.addCommand('zp', 'paste-cell', 'cursorCol.setValuesTyped([cursorRow], vd.memory.clipval) if vd.memory.clipval else vd.warning("nothing to paste from clipval")', 'set contents of current cell to last clipboard value')

Sheet.addCommand('d', 'delete-row', 'delete_row(cursorRowIndex); defer and cursorDown(1)', 'delete current row')
Sheet.addCommand('gd', 'delete-selected', 'deleteSelected()', 'delete selected rows')
Sheet.addCommand('zd', 'delete-cell', 'cursorCol.setValues([cursorRow], options.null_value)', 'delete current cell (set to None)')
Sheet.addCommand('gzd', 'delete-cells', 'cursorCol.setValues(onlySelectedRows, options.null_value)', 'delete contents of current column for selected rows (set to None)')

Sheet.bindkey('BUTTON2_PRESSED', 'go-mouse')
Sheet.addCommand('BUTTON2_RELEASED', 'syspaste-cells', 'pasteFromClipboard(visibleCols[cursorVisibleColIndex:], rows[cursorRowIndex:])', 'paste from system clipboard to region starting at cursor')
Sheet.bindkey('BUTTON2_CLICKED', 'go-mouse')
Sheet.bindkey('zP', 'syspaste-cells')
Sheet.addCommand('gzP', 'syspaste-cells-selected', 'pasteFromClipboard(visibleCols[cursorVisibleColIndex:], someSelectedRows)', 'paste from system clipboard to selected cells')

Sheet.addCommand('gzy', 'copy-cells', 'copyCells(cursorCol, onlySelectedRows)', 'yank (copy) contents of current column for selected rows to clipboard')
Sheet.addCommand('gzp', 'setcol-clipboard', 'setColClipboard()', 'set cells of current column for selected rows to last clipboard value')

Sheet.addCommand('Y', 'syscopy-row', 'syscopyCells(visibleCols, [cursorRow])', 'yank (copy) current row to system clipboard (using options.clipboard_copy_cmd)')

Sheet.addCommand('gY', 'syscopy-selected', 'syscopyCells(visibleCols, onlySelectedRows)', 'yank (copy) selected rows to system clipboard (using options.clipboard_copy_cmd)')
Sheet.addCommand('zY', 'syscopy-cell', 'syscopyValue(cursorDisplay)', 'yank (copy) current cell to system clipboard (using options.clipboard_copy_cmd)')
Sheet.addCommand('', 'syscopy-colname', 'syscopyValue(cursorCol.name)', 'yank (copy) current column header to system clipboard (using options.clipboard_copy_cmd)')
Sheet.addCommand('gzY', 'syscopy-cells', 'syscopyCells([cursorCol], onlySelectedRows, filetype="txt")', 'yank (copy) contents of current column from selected rows to system clipboard (using options.clipboard_copy_cmd')

Sheet.addCommand('x', 'cut-row', 'copyRows([sheet.delete_row(cursorRowIndex)]); defer and cursorDown(1)', 'delete (cut) current row and move it to clipboard')
Sheet.addCommand('gx', 'cut-selected', 'copyRows(onlySelectedRows); deleteSelected()', 'delete (cut) selected rows and move them to clipboard')
Sheet.addCommand('zx', 'cut-cell', 'copyCells(cursorCol, [cursorRow]); cursorCol.setValues([cursorRow], None)', 'delete (cut) current cell and move it to clipboard')
Sheet.addCommand('gzx', 'cut-cells', 'copyCells(cursorCol, onlySelectedRows); cursorCol.setValues(onlySelectedRows, None)', 'delete (cut) contents of current column for selected rows and move them to clipboard')


Sheet.bindkey('KEY_DC', 'delete-cell'),
Sheet.bindkey('gKEY_DC', 'delete-cells'),

vd.addMenuItems('''
    Edit > Delete > current row > delete-row
    Edit > Delete > current cell > delete-cell
    Edit > Delete > selected rows > delete-selected
    Edit > Delete > selected cells > delete-cells
    Edit > Copy > current cell > copy-cell
    Edit > Copy > current row > copy-row
    Edit > Copy > selected cells > copy-cells
    Edit > Copy > selected rows > copy-selected
    Edit > Copy > to system clipboard > current cell > syscopy-cell
    Edit > Copy > to system clipboard > current row > syscopy-row
    Edit > Copy > to system clipboard > selected cells > syscopy-cells
    Edit > Copy > to system clipboard > selected rows > syscopy-selected
    Edit > Cut > current row > cut-row
    Edit > Cut > selected cells > cut-selected
    Edit > Cut > current cell > cut-cell
    Edit > Paste > row after > paste-after
    Edit > Paste > row before > paste-before
    Edit > Paste > into selected cells > setcol-clipboard
    Edit > Paste > into current cell > paste-cell
    Edit > Paste > from system clipboard > cells at cursor > syspaste-cells
    Edit > Paste > from system clipboard > selected cells > syspaste-cells-selected
''')



'''
Enables two functions:
1. options.disp_scroll_context: the minimum number of lines to keep visible above and below the cursor when scrolling
2. toggle-scrollfix: Lock the cursor to a particular row number in the view

Allows the user to fix the position on the page where they would like the
cursor to "stick". Helps to provide context about surrounding rows when
near the top and bottom of the page.

Usage: scroll a few lines down, `toggle-scrollfix' and scroll again!

NOTE:
    - disables scroll-middle command (zz)
'''

__author__ = 'Geekscrapy'
__version__ = '1.2'

from visidata import vd, Sheet

vd.option('disp_scroll_context', 0, 'minimum number of lines to keep visible above/below cursor when scrolling')
vd.optalias('disp_scrolloff', 'disp_scroll_context')

@Sheet.after
def checkCursor(sheet):
    nctx = sheet.options.disp_scroll_context
    if nctx:
        if nctx >= int(sheet.nScreenRows/2):
            if sheet.cursorRowIndex+1 > int(sheet.nScreenRows/2):
                sheet.topRowIndex = sheet.cursorRowIndex - int(sheet.nScreenRows/2)
        else:
            if sheet.cursorRowIndex-sheet.topRowIndex < nctx:
                sheet.topRowIndex = max(0, sheet.cursorRowIndex-nctx)

            if sheet.bottomRowIndex-sheet.cursorRowIndex < nctx:
                sheet.bottomRowIndex = sheet.cursorRowIndex+nctx

    nfix = getattr(sheet, 'disp_scrollfix', 0)
    if nfix > 0:
        cursorViewIndex = sheet.cursorRowIndex - sheet.topRowIndex
        if cursorViewIndex > nfix:
            sheet.topRowIndex += abs(nfix - cursorViewIndex)
        if cursorViewIndex < nfix and sheet.topRowIndex > 0:
            sheet.topRowIndex -= abs(nfix - cursorViewIndex)


@Sheet.api
def toggle_scrollfix(sheet):
    if getattr(sheet, 'disp_scrollfix', -1) >= 0:
        sheet.disp_scrollfix = -1
        vd.status("cursor unlocked")
    else:
        sheet.disp_scrollfix = sheet.cursorRowIndex - sheet.topRowIndex
        vd.status(f"cursor locked to screen row {sheet.disp_scrollfix}")


Sheet.addCommand('', 'toggle-scrollfix', 'toggle_scrollfix()', helpstr='toggle cursor lock to current screen row')


vd.addMenuItems('''View > Toggle display > lock cursor to screen row > toggle-scrollfix''')



'Light-mode theme using 256-colors.'

from visidata import vd

vd.themes['light'] = dict(
        color_default      = 'black on white',  # the default fg and bg colors
        color_default_hdr  = 'bold black on white',
        color_bottom_hdr   = 'underline black on white',
        color_current_col  = 'bold on white',
        color_column_sep   = 'black on white',
        color_key_col      = '20 blue',   # color of key columns
        color_edit_cell    = '234 black',     # cell color to use when editing cell
        color_selected_row = '164 magenta',  # color of selected rows
        color_selected_col = 'bold on white',
        color_note_row     = '164 magenta',  # color of row note on left edge
        color_note_type    = '88 red',  # color of cell note for non-str types in anytype columns
        color_warning      = '202 11 yellow',
        color_add_pending  = '34 green',
        color_change_pending  = '166 yellow',
        plot_colors = '20 red magenta black 28 88 94 99 106'
)




import re
from visidata import vd, VisiData, BaseSheet, Sheet, Column, Progress, asyncthread, rotateRange

VisiData.init('searchContext', dict) # [(regex, columns, backward)] -> kwargs from previous search

vd.help_regex_flags = '''# Regex Flags Help
- `A` (ASCII) ASCII-only matching (not unicode)
- `I` (IGNORECASE): case-insensitive matching
- `M` (MULTILINE): `^` and `$` match after/before newlines
- `S` (DOTALL): `.` match any character at all, including newline
- `X` (VERBOSE): allow verbose regex
'''


@VisiData.api
@asyncthread
def moveRegex(vd, sheet, *args, **kwargs):
    list(vd.searchRegex(sheet, *args, moveCursor=True, **kwargs))


# kwargs: regex=None, columns=None, backward=False
@VisiData.api
def searchRegex(vd, sheet, moveCursor=False, reverse=False, regex_flags=None, **kwargs):
        'Set row index if moveCursor, otherwise return list of row indexes.'
        def findMatchingColumn(sheet, row, columns, func):
            'Find column for which func matches the displayed value in this row'
            for c in columns:
                if func(c.getDisplayValue(row)):
                    return c

        vd.searchContext.update(kwargs)

        regex = kwargs.get("regex")
        if regex:
            if regex_flags is None:
                regex_flags = sheet.options.regex_flags  # regex_flags defined in features.regex
            flagbits = sum(getattr(re, f.upper()) for f in regex_flags)
            try:
                compiled_re = re.compile(regex, flagbits)
                vd.searchContext["regex"] = compiled_re
            except re.error as e:
                vd.searchContext["regex"] = None  # make future calls to search-next fail
                vd.error('invalid regex: %s' % e.msg)

        regex = vd.searchContext.get("regex") or vd.fail("no regex")

        columns = vd.searchContext.get("columns")
        if columns == "cursorCol":
            columns = [sheet.cursorCol]
        elif columns == "visibleCols":
            columns = tuple(sheet.visibleCols)
        elif isinstance(columns, Column):
            columns = [columns]

        if not columns:
            vd.error('bad columns')

        searchBackward = vd.searchContext.get("backward")
        if reverse:
            searchBackward = not searchBackward

        matchingRowIndexes = 0
        for rowidx in rotateRange(len(sheet.rows), sheet.cursorRowIndex, reverse=searchBackward):
            c = findMatchingColumn(sheet, sheet.rows[rowidx], columns, regex.search)
            if c:
                if moveCursor:
                    sheet.cursorRowIndex = rowidx
                    sheet.cursorVisibleColIndex = sheet.visibleCols.index(c)
                    return
                else:
                    matchingRowIndexes += 1
                    yield rowidx

        if kwargs.get('printStatus', True):
            vd.status('%s matches for /%s/' % (matchingRowIndexes, regex.pattern))


@Sheet.api
def searchInputRegex(sheet, action:str, columns:str='cursorCol'):
    r = vd.inputMultiple(regex=dict(prompt=f"{action} regex: ", type="regex", defaultLast=True, help=vd.help_regex),
                         flags=dict(prompt="regex flags: ", type="regex_flags", value=sheet.options.regex_flags, help=vd.help_regex_flags))

    return vd.searchRegex(sheet, regex=r['regex'], regex_flags=r['flags'], columns=columns)

@Sheet.api
def moveInputRegex(sheet, action:str, type="regex", **kwargs):
    r = vd.inputMultiple(regex=dict(prompt=f"{action} regex: ", type=type, defaultLast=True, help=vd.help_regex),
                         flags=dict(prompt="regex flags: ", type="regex_flags", value=sheet.options.regex_flags, help=vd.help_regex_flags))

    return vd.moveRegex(sheet, regex=r['regex'], regex_flags=r['flags'], **kwargs)

@Sheet.api
@asyncthread
def search_expr(sheet, expr, reverse=False, curcol=None):
    for i in rotateRange(len(sheet.rows), sheet.cursorRowIndex, reverse=reverse):
        try:
            if sheet.evalExpr(expr, sheet.rows[i], curcol=curcol):
                sheet.cursorRowIndex=i
                return
        except Exception as e:
            vd.exceptionCaught(e)

    vd.fail(f'no {sheet.rowtype} where {expr}')


Sheet.addCommand('r', 'search-keys', 'tmp=cursorVisibleColIndex; moveInputRegex("row key", type="regex-row", columns=keyCols or [visibleCols[0]]); sheet.cursorVisibleColIndex=tmp', 'go to next row with key matching regex')
Sheet.addCommand('/', 'search-col', 'moveInputRegex("search", columns="cursorCol", backward=False)', 'search for regex forwards in current column')
Sheet.addCommand('?', 'searchr-col', 'moveInputRegex("reverse search", columns="cursorCol", backward=True)', 'search for regex backwards in current column')
Sheet.addCommand('n', 'search-next', 'vd.moveRegex(sheet, reverse=False)', 'go to next match from last regex search')
Sheet.addCommand('N', 'searchr-next', 'vd.moveRegex(sheet, reverse=True)', 'go to previous match from last regex search')

Sheet.addCommand('g/', 'search-cols', 'moveInputRegex("g/", backward=False, columns="visibleCols")', 'search for regex forwards over all visible columns')
Sheet.addCommand('g?', 'searchr-cols', 'moveInputRegex("g?", backward=True, columns="visibleCols")', 'search for regex backwards over all visible columns')
Sheet.addCommand('z/', 'search-expr', 'search_expr(inputExpr("search by expr: ") or fail("no expr"), curcol=cursorCol)', 'search by Python expression forwards in current column (with column names as variables)')
Sheet.addCommand('z?', 'searchr-expr', 'search_expr(inputExpr("searchr by expr: ") or fail("no expr"), curcol=cursorCol, reverse=True)', 'search by Python expression backwards in current column (with column names as variables)')

vd.addMenuItems('''
    View > Search > current column > search-col
    View > Search > visible columns > search-cols
    View > Search > key columns > search-keys
    View > Search > by Python expr > search-expr
    View > Search > again > search-next
    View > Search backward > current column > searchr-col
    View > Search backward > visible columns > searchr-cols
    View > Search backward > by Python expr > searchr-expr
    View > Search backward > again > searchr-next
''')



'''# RedditSheet

- [:keystrokes]Ctrl+O[/] to open a browser tab to [:code]{sheet.cursorRow.display_name_prefixed}[/]
- [:keystrokes]g Ctrl+O[/] to open browser windows for {sheet.nSelectedRows} selected subreddits

- [:keystrokes]Enter[/] to open sheet with top ~1000 submissions for [:code]{sheet.cursorRow.display_name_prefixed}[/]
- [:keystrokes]g Enter[/] to open sheet with top ~1000 submissions for {sheet.nSelectedRows} selected subreddits

- [:keystrokes]ga[/] to append more subreddits matching input by name or description
'''

import visidata
from visidata import vd, VisiData, Sheet, AttrColumn, asyncthread, ENTER, anytype, date


vd.option('reddit_client_id', '', 'client_id for reddit api')
vd.option('reddit_client_secret', '', 'client_secret for reddit api')
vd.option('reddit_user_agent', visidata.__version__, 'user_agent for reddit api')


@VisiData.api
def open_reddit(vd, p):
    vd.importExternal('praw')
    vd.enable_requests_cache()

    if not vd.options.reddit_client_id:
        return RedditGuide('reddit_guide')

    if p.given.startswith('r/') or p.given.startswith('/r/'):
        return SubredditSheet(p.base_stem, source=p.base_stem.split('+'), search=(p.given[0]=='/'))

    if p.given.startswith('u/') or p.given.startswith('/u/'):
        return RedditorsSheet(p.base_stem, source=p.base_stem.split('+'), search=(p.given[0]=='/'))

    return SubredditSheet(p.base_stem, source=p)

vd.new_reddit = vd.open_reddit

@VisiData.cached_property
def reddit(vd):
    import praw
    return praw.Reddit(check_for_updates=False, **vd.options.getall('reddit_'))


subreddit_hidden_attrs='''
name #accounts_active accounts_active_is_fuzzed advertiser_category
all_original_content allow_chat_post_creation allow_discovery
allow_galleries allow_images allow_polls allow_predictions
allow_predictions_tournament allow_videogifs allow_videos
banner_background_color banner_background_image banner_img banner_size
can_assign_link_flair can_assign_user_flair collapse_deleted_comments
comment_score_hide_mins community_icon community_reviewed @created
@created_utc description_html disable_contributor_requests
display_name display_name_prefixed emoji emojis_custom_size
emojis_enabled filters free_form_reports fullname has_menu_widget
header_img header_size header_title hide_ads icon_img icon_size
is_chat_post_feature_enabled is_crosspostable_subreddit
is_enrolled_in_new_modmail key_color lang link_flair_enabled
link_flair_position mobile_banner_image mod notification_level
original_content_tag_enabled over18 prediction_leaderboard_entry_type
primary_color public_description public_description_html public_traffic
quaran quarantine restrict_commenting restrict_posting show_media
show_media_preview spoilers_enabled submission_type submit_link_label
submit_text submit_text_html submit_text_label suggested_comment_sort
user_can_flair_in_sr user_flair_background_color user_flair_css_class
user_flair_enabled_in_sr user_flair_position user_flair_richtext
user_flair_template_id user_flair_text user_flair_text_color
user_flair_type user_has_favorited user_is_banned user_is_contributor
user_is_moderator user_is_muted user_is_subscriber user_sr_flair_enabled
user_sr_theme_enabled #videostream_links_count whitelist_status widgets
wiki wiki_enabled wls
'''

post_hidden_attrs='''
all_awardings allow_live_comments @approved_at_utc approved_by archived
author_flair_background_color author_flair_css_class author_flair_richtext
author_flair_template_id author_flair_text author_flair_text_color
author_flair_type author_fullname author_patreon_flair author_premium
awarders @banned_at_utc banned_by can_gild can_mod_post category clicked
comment_limit comment_sort content_categories contest_mode @created_utc
discussion_type distinguished domain edited flair fullname gilded
gildings hidden hide_score is_crosspostable is_meta is_original_content
is_reddit_media_domain is_robot_indexable is_self is_video likes
link_flair_background_color link_flair_css_class link_flair_richtext
link_flair_text link_flair_text_color link_flair_type locked media
media_embed media_only mod mod_note mod_reason_by mod_reason_title
mod_reports name no_follow num_crossposts num_duplicates num_reports
over_18 parent_whitelist_status permalink pinned pwls quarantine
removal_reason removed_by removed_by_category report_reasons saved
score secure_media secure_media_embed selftext_html send_replies
shortlink spoiler stickied subreddit_id subreddit_name_prefixed
subreddit_subscribers subreddit_type suggested_sort thumbnail
thumbnail_height thumbnail_width top_awarded_type total_awards_received
treatment_tags upvote_ratio user_reports #view_count visited
whitelist_status wls
'''

comment_hidden_attrs='''
all_awardings @approved_at_utc approved_by archived associated_award
author_flair_background_color author_flair_css_class author_flair_richtext
author_flair_template_id author_flair_text author_flair_text_color
author_flair_type author_fullname author_patreon_flair author_premium
awarders @banned_at_utc banned_by body_html can_gild can_mod_post
collapsed collapsed_because_crowd_control collapsed_reason comment_type
controversiality @created_utc distinguished fullname gilded gildings
is_root is_submitter likes link_id locked mod mod_note mod_reason_by
mod_reason_title mod_reports name no_follow num_reports parent_id
permalink removal_reason report_reasons saved #score #score_hidden
send_replies stickied submission subreddit_id subreddit_name_prefixed
subreddit_type top_awarded_type total_awards_received treatment_tags
user_reports
'''

redditor_hidden_attrs='''
#awardee_karma #awarder_karma @created @created_utc
fullname has_subscribed has_verified_email hide_from_robots icon_img id
is_employee is_friend is_gold is_mod pref_show_snoovatar
snoovatar_img snoovatar_size stream #total_karma verified
subreddit.banner_img subreddit.name subreddit.over_18 subreddit.public_description #subreddit.subscribers subreddit.title
'''

def hiddenCols(hidden_attrs):
    coltypes = { t.icon:t.typetype for t in vd.typemap.values() if not t.icon.isalpha() }
    for attr in hidden_attrs.split():
        coltype = anytype
        if attr[0] in coltypes:
            coltype = coltypes.get(attr[0])
            attr = attr[1:]
        yield AttrColumn(attr, type=coltype, width=0)


class SubredditSheet(Sheet):
    guide = __doc__
    # source is a text list of subreddits
    rowtype = 'subreddits'  # rowdef: praw.Subreddit
    nKeys=1
    search=False
    columns = [
        AttrColumn('display_name_prefixed', width=15),
        AttrColumn('active_user_count', type=int),
        AttrColumn('subscribers', type=int),
        AttrColumn('subreddit_type'),
        AttrColumn('title'),
        AttrColumn('description', width=50),
        AttrColumn('url', width=10),
    ] + list(hiddenCols(subreddit_hidden_attrs))

    def iterload(self):
        for name in self.source:
            name = name.strip()
            if self.search:
                yield from vd.reddit.subreddits.search(name)
            else:
                try:
                    r = vd.reddit.subreddit(name)
                    r.display_name_prefixed
                    yield r
                except Exception as e:
                    vd.exceptionCaught(e)

    def openRow(self, row):
        return RedditSubmissions(row.display_name_prefixed, source=row)

    def openRows(self, rows):
        comboname = '+'.join(row.display_name for row in rows)
        return RedditSubmissions(comboname, source=vd.reddit.subreddit(comboname))


class RedditorsSheet(Sheet):
    # source is a text list of usernames
    rowtype = 'redditors'  # rowdef: praw.Subreddit
    nKeys=1
    columns = [
        AttrColumn('name', width=15),
        AttrColumn('comment_karma', type=int),
        AttrColumn('link_karma', type=int),
        AttrColumn('comments'),
        AttrColumn('submissions'),
    ] + list(hiddenCols(redditor_hidden_attrs))

    def iterload(self):
        for name in self.source:
            if self.search:
                yield from vd.reddit.redditors.popular(name)
            else:
                yield vd.reddit.redditor(name)

    def openRow(self, row):
        return RedditSubmissions(row.fullname, source=row.submissions)

    def openRows(self, rows):
        comboname = '+'.join(row.name for row in rows)
        return RedditSubmissions(comboname, source=vd.reddit.redditor(comboname).submissions)


class RedditSubmissions(Sheet):
    guide = '''# Reddit Submissions

  [:keys]Enter[/] to open sheet with comments for the current post
  [:keys]ga[/] to add posts in this subreddit matching input'''

    # source=ListingGenerator
    rowtype='reddit posts' # rowdef: praw.Submission
    nKeys=2
    columns = [
        AttrColumn('subreddit'),
        AttrColumn('id', width=0),
        AttrColumn('created', width=12, type=date),
        AttrColumn('author'),
        AttrColumn('ups', width=8, type=int),
        AttrColumn('downs', width=8, type=int),
        AttrColumn('num_comments', width=8, type=int),
        AttrColumn('title', width=50),
        AttrColumn('selftext', width=60),
        AttrColumn('url'),
        AttrColumn('comments', width=0),
    ] + list(hiddenCols(post_hidden_attrs))

    def iterload(self):
        kind = 'new' # 'top'
        f = getattr(self.source, kind, None)
        if f:
            yield from f(limit=10000)

    def openRow(self, row):
        return RedditComments(row.id, source=row.comments.list())


class RedditComments(Sheet):
    # source=list of comments
    rowtype='comments' # rowdef: praw.Comment
    nKeys=2
    columns=[
        AttrColumn('subreddit', width=0),
        AttrColumn('id', width=0),
        AttrColumn('ups', width=4, type=int),
        AttrColumn('downs', width=4, type=int),
        AttrColumn('replies', type=list),
        AttrColumn('created', type=date),
        AttrColumn('author'),
        AttrColumn('depth', type=int),
        AttrColumn('body', width=60),
        AttrColumn('edited', width=0),
    ] + list(hiddenCols(comment_hidden_attrs))

    def iterload(self):
        yield from self.source

    def openRow(self, row):
        return RedditComments(row.id, source=row.replies)


class RedditGuide(RedditSubmissions):
    guide = '''# Authenticate Reddit
The Reddit API must be configured before use.

1. Login to Reddit and go to [:underline]https://www.reddit.com/prefs/apps[/].
2. Create a "script" app. (Use "[:underline]http://localhost:8000[/]" for the redirect uri)
3. Add credentials to visidatarc:

    options.reddit_client_id = '...'      # below the description in the upper left
    options.reddit_client_secret = '...'

## Use [:code]reddit[/] filetype for subreddits or users

Multiple may be specified, joined with "+".

    vd r/commandline.reddit
    vd u/gallowboob.reddit
    vd r/rust+golang+python.reddit
    vd u/spez+kn0thing.reddit
'''

@SubredditSheet.api
@asyncthread
def addRowsFromQuery(sheet, q):
    for r in vd.reddit.subreddits.search(q):
        sheet.addRow(r, index=sheet.cursorRowIndex+1)


@RedditSubmissions.api
@asyncthread
def addRowsFromQuery(sheet, q):
    for r in sheet.source.search(q, limit=None):
        sheet.addRow(r, index=sheet.cursorRowIndex+1)


@VisiData.api
def sysopen_subreddits(vd, *subreddits):
    url = "https://www.reddit.com/r/"+"+".join(subreddits)
    vd.launchBrowser(url)


SubredditSheet.addCommand('^O', 'sysopen-subreddit', 'sysopen_subreddits(cursorRow.display_name)', 'open browser window with subreddit')
SubredditSheet.addCommand('g^O', 'sysopen-subreddits', 'sysopen_subreddits(*(row.display_name for row in selectedRows))', 'open browser window with messages from selected subreddits')
SubredditSheet.addCommand('g'+ENTER, 'open-subreddits', 'vd.push(openRows(selectedRows))', 'open sheet with top ~1000 submissions for each selected subreddit')
SubredditSheet.addCommand('ga', 'add-subreddits-match', 'addRowsFromQuery(input("add subreddits matching: "))', 'add subreddits matching input by name or description')
RedditSubmissions.addCommand('ga', 'add-submissions-match', 'addRowsFromQuery(input("add posts matching: "))', 'add posts in this subreddit matching input')

vd.addMenuItems('''
    File > Reddit > open selected subreddits > open-subreddits
    File > Reddit > add > matching subreddits > add-subreddits-match
    File > Reddit > add > matching submissions > add-submissions-match
    File > Reddit > open in browser > subreddit in current row > sysopen-subreddit
    File > Reddit > open in browser > selected subreddits > sysopen-subreddits
''')



import functools

from visidata import vd, VisiData, Column, ItemColumn, date, RowColorizer, asyncthread, Progress, AttrDict

from .gitsheet import GitSheet


@VisiData.api
def git_log(vd, p, *args):
    return GitLogSheet('git-log', source=p, gitargs=args)

# rowdef: (commit_hash, refnames, author, author_date, body, notes)
class GitLogSheet(GitSheet):
    guide = '''
# git log {sheet.gitargstr}
{sheet.cursorRow.message}
    '''
    GIT_LOG_FORMAT = ['%H', '%D', '%an <%ae>', '%ai', '%B', '%N']
    rowtype = 'commits'  # rowdef: AttrDict
    defer = True
    savesToSource = True
    columns = [
        ItemColumn('commitid', width=8),
        ItemColumn('refnames', width=12),
        ItemColumn('message', type=str.strip, setter=lambda c,r,v: c.sheet.git('commit --amend --no-edit --quiet --message', v), width=50),
        ItemColumn('author', setter=lambda c,r,v: c.sheet.git('commit --amend --no-edit --quiet --author', v)),
        ItemColumn('author_date', type=date, setter=lambda c,r,v: c.sheet.git('commit --amend --no-edit --quiet --date', v)),
        ItemColumn('notes', setter=lambda c,r,v: c.sheet.git('notes add --force --message', v, r.commitid)),
    ]
    colorizers = [
            RowColorizer(5, 'color_vgit_unpushed', lambda s,c,r,v: r and not s.inRemoteBranch(r.commitid)),
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @functools.lru_cache()
    def inRemoteBranch(self, commitid):
        return self.git_all('branch -r --contains', commitid, _ok_code=[0, 1])

    def iterload(self):
        lines = self.git_iter('log --no-color -z', '--pretty=format:' + '%x1f'.join(self.GIT_LOG_FORMAT), *self.gitargs)
        for record in Progress(tuple(lines)):
            r = record.split('\x1f')
            yield AttrDict(
                commitid=r[0],
                refnames=r[1],
                author=r[2],
                author_date=r[3],
                message=r[4],
                notes=r[5],
            )

    def openRow(self, row):
        'open this commit'
        return getCommitSheet(row[0][:7], self, row[0])

    @asyncthread
    def commit(self, path, adds, mods, dels):

        assert not adds
        assert not dels

        for row, rowmods in mods.values():
            for col, val in rowmods.values():
                vd.callNoExceptions(col.putValue, row, val)

        self.reload()
        self.resetDeferredCommit()


GitLogSheet.addCommand(None, 'delete-row', 'error("delete is not supported")')
GitLogSheet.addCommand(None, 'add-row', 'error("commits cannot be added")')
#GitLogSheet.addCommand('x', 'git-pick', 'git("cherry-pick", cursorRow.commitid)', 'cherry-pick this commit onto current branch')
#GitLogSheet.addCommand('r', 'git-reset-here', 'git("update-ref", "refs/heads/"+source, cursorRow[0])', 'reset this branch to this commit')

GitSheet.addCommand('', 'git-log', 'vd.push(git_log(gitRootPath, branch))', 'push log of current branch')

vd.addMenuItems('''
    Git > Open > log > git-log
''')



# requires shell tools: ping traceroute
#   these tools aren't inlined as they require root privs

import re
import time
from statistics import mean

from visidata import vd, VisiData, BaseSheet, Sheet, Column, AttrColumn, Progress

@VisiData.api
def new_ping(vd, p):
    'Open a sheet with the round-trip time for each hop along the path to the given host.'
    return makePingStats(p.given)

# open_ping() is called when a local file exists matching the path p, otherwise new_ping() is called
vd.open_ping = vd.new_ping

def makePingStats(ip):
    pingsheet = PingSheet(ip, source=ip)
    return PingStatsSheet("traceroute_"+pingsheet.name, source=pingsheet)

class PingStatsSheet(Sheet):
    help='''# ping/traceroute
This sheet runs `traceroute` to generate intermediate hops, then runs `ping` against each hop N times to get the `avg_ms` and `max_ms` ping time.

- {help.commands.open_source} (for raw ping data)
'''
    rowtype='hosts' # rowdef: PingColumn
    columns = [
        Column('hop', type=int, width=5, getter=lambda col,r: col.sheet.source.sources.index(r.ip)+1),
        Column('hostname', getter=lambda col,r: r.name, width=40),
        Column('ip',  getter=lambda col,r: r.ip, width=17),
        Column('avg_ms', type=float, fmtstr='%.1f', getter=lambda col,r: mean(r.getValues(col.sheet.source.rows))),
        Column('max_ms', type=float, fmtstr='%.1f', getter=lambda col,r: max(r.getValues(col.sheet.source.rows))),
        Column('count',  type=int, getter=lambda col,r: len(list(r.getValues(col.sheet.source.rows)))),
    ]
    def loader(self):
        sheet = self.source
        vd.sync(sheet.ensureLoaded())

        self.rows = sheet.columns


def PingColumn(name, ip):
    return Column(name, ip=ip, getter=lambda c,r: r.get(c.ip), type=float, fmtstr='%0.1f')


class PingSheet(Sheet):
    rowtype = 'pings'  # rowdef: {'time':time.time(), 'hostname': pingtime}
    def __init__(self, *names, source=None, **kwargs):
        super().__init__(*names, source=source, **kwargs)
        self.sources = [source]

    def ping_response(self, row, ip, data):
        for line in data.splitlines():
            m = re.search(r'from ([0-9\.]+).* time=(.*) ms', line)
            if m:
                row[ip] = m.group(2)
#                assert ip == m.group(1)

    def traceroute_response(self, ip, data):
        for line in data.splitlines():
            m = re.search(r'(\d+)  (\S+) \((\S+)\)  (.*) ms', line)
            if m:
                ttl, hostname, inner_ip, latency_ms = m.groups()
                self.routes[ip][int(ttl)-1] = inner_ip
                if inner_ip == ip:
                    return  # traceroute work is done, let ping do the rest

                if inner_ip not in self.sources:
                    self.sources.insert(-1, inner_ip)
                    self.addColumn(PingColumn(hostname, inner_ip), index=-1)
                    self.send_trace(ip, int(ttl)+1)  # get next hop
                break

    def ping_error(self, ip, data):
        if ip in self.sources:
            self.sources.remove(ip)
            vd.warning("%s removed: %s" % (ip, data.rstrip()))

    def update_traces(self, row, ip):
        rtes = self.routes.get(ip)
        if rtes is None:
            rtes = []
            self.routes[ip] = rtes

        if ip in rtes:
            return

        for i, inner_ip in enumerate(rtes):
            if inner_ip is None:
                self.send_trace(ip, i+1)

        self.send_trace(ip, len(rtes)+1)  # and one more for the road

    def send_trace(self, ip, n):
        sh = vd.importExternal('sh')

        while n > len(self.routes[ip]):
            self.routes[ip].append(None)

        sh.traceroute('--sim-queries=1',
                      '--queries=1',
                      '--first=%s' % n,
                      '--max-hops=%s' % n,
                      ip,
                      _bg=True, _bg_exc=False,
                      _out=lambda data,self=self,ip=ip,a=1: self.traceroute_response(ip, data),
                      _err=lambda data,self=self,ip=ip,a=1: self.ping_error(ip, data))

    def iterload(self):
        sh = vd.importExternal('sh')

        self.stop = False
        self.start_time = time.time()
        self.columns.clear()

        for ip in self.sources:
            self.addColumn(PingColumn(ip, ip))

        self.routes = {}
        self.rows = []
        pings_sent = {}
        ping_count = self.options.ping_count
        with Progress(total=ping_count*len(self.sources), gerund='pinging') as prog:
          while not self.stop:
            r = {'time':time.time()}
            yield r

            npings = 0  # this loop
            for ip in self.sources:
                self.update_traces(r, ip)
                pings_sent[ip] = pings_sent.get(ip, 0)+1
                if ping_count and pings_sent[ip] <= ping_count:
                    sh.ping('-c', '1', ip, _bg=True, _bg_exc=False, _timeout=30,
                            _out=lambda data,self=self,r=r,ip=ip: self.ping_response(r,ip,data),
                            _err=lambda data,self=self,ip=ip,a=1: self.ping_error(ip, data))
                    npings += 1
                else:
                    r[ip] = None

            if npings == 0:
                vd.status('no more pings to send')
                break

            time.sleep(self.options.ping_interval)


vd.option('ping_count', 3, 'send this many pings to each host', sheettype=PingSheet)
vd.option('ping_interval', 0.1, 'wait between ping rounds, in seconds', sheettype=PingSheet)

PingSheet.options.null_value = False
BaseSheet.addCommand('', 'open-ping', 'vd.push(makePingStats(vd.input("ping: ", type="hostip")))', 'open sheet to ping input IP Address')

vd.addGlobals(
    makePingStats=makePingStats,
)

vd.addMenuItems('''
    System > Ping IP/hostname > open-ping
''')



# VisiData uses Python native int, float, str, and adds simple anytype.

import locale
from visidata import vd, VisiData

vd.help_float_fmt = '''
- fmt starting with `'%'` (like `%0.2f`) will use [:onclick https://docs.python.org/3.6/library/locale.html#locale.format_string]locale.format_string[/]
- other fmt (like `{:.02f}` is passed to Python [:onclick https://docs.python.org/3/library/string.html#custom-string-formatting)]string.format][/]
'''

vd.help_int_fmt = '''
- fmt starting with `'%'` (like `%04d`) will use [:onclick https://docs.python.org/3.6/library/locale.html#locale.format_string]locale.format_string[/]
- other fmt (like `{:4d}` is passed to Python [:onclick https://docs.python.org/3/library/string.html#custom-string-formatting)]string.format[/]
'''

vd.option('disp_float_fmt', '{:.02f}', 'default fmtstr to format float values', replay=True, help=vd.help_float_fmt)
vd.option('disp_int_fmt', '{:d}', 'default fmtstr to format int values', replay=True, help=vd.help_int_fmt)


vd.numericTypes = [int,float]

# VisiDataType .typetype are e.g. int, float, str, and used internally in these ways:
#
#    o = typetype(val)   # for interpreting raw value
#    o = typetype(str)   # for conversion from string (when setting)
#    o = typetype()      # for default value to be used when conversion fails
#
# The resulting object o must be orderable and convertible to a string for display and certain outputs (like csv).
#
# .icon is a single character that appears in the notes field of cells and column headers.
# .formatter(fmtstr, typedvalue) returns a string of the formatted typedvalue according to fmtstr.
# .fmtstr is the default fmtstr passed to .formatter.

def anytype(r=None):
    'minimalist "any" passthrough type'
    return r
anytype.__name__ = ''


@VisiData.global_api
def numericFormatter(vd, fmtstr, typedval):
    try:
        fmtstr = fmtstr or vd.options['disp_'+type(typedval).__name__+'_fmt']
        if fmtstr[0] == '%':
            return locale.format_string(fmtstr, typedval, grouping=False)
        else:
            return fmtstr.format(typedval)
    except ValueError:
        return str(typedval)


@VisiData.api
def numericType(vd, icon='', fmtstr='', formatter=vd.numericFormatter):
    '''Decorator for numeric types.'''
    def _decorator(f):
        vd.addType(f, icon=icon, fmtstr=fmtstr, formatter=formatter, name=f.__name__)
        vd.numericTypes.append(f)
        return f
    return _decorator


class VisiDataType:
    'Register *typetype* in the typemap.'
    def __init__(self, typetype=None, icon=None, fmtstr='', formatter=vd.numericFormatter, key='', name=None):
        self.typetype = typetype or anytype # int or float or other constructor
        self.name = name or getattr(typetype, '__name__', str(typetype))
        self.icon = icon      # show in rightmost char of column
        self.fmtstr = fmtstr
        self.formatter = formatter
        self.key = key

@VisiData.api
def addType(vd, typetype=None, icon=None, fmtstr='', formatter=vd.numericFormatter, key='', name=None):
    '''Add type to type map.

    - *typetype*: actual type class *TYPE* above
    - *icon*: unicode character in column header
    - *fmtstr*: format string to use if fmtstr not given
    - *formatter*: formatting function to call as ``formatter(fmtstr, typedvalue)``
    '''
    t = VisiDataType(typetype=typetype, icon=icon, fmtstr=fmtstr, formatter=formatter, key=key, name=name)
    if typetype:
        vd.typemap[typetype] = t
        if name:
            vd.addGlobals({name: typetype})
    return t

vdtype = vd.addType


# typemap [vtype] -> VisiDataType
vd.typemap = {}

@VisiData.api
def getType(vd, typetype):
    return vd.typemap.get(typetype) or VisiDataType()

vdtype(None, '∅', name='none')
vdtype(anytype, '', formatter=lambda _,v: str(v))
vdtype(str, '~', formatter=lambda _,v: v)
vdtype(int, '#')
vdtype(float, '%')
vdtype(dict, '')
vdtype(list, '')

@VisiData.api
def isNumeric(vd, col):
    return col.type in vd.numericTypes

def deduceType(v):
    if isinstance(v, (float, int)):
        return type(v)
    else:
        return anytype
##

@vd.numericType('%')
def floatlocale(*args):
    'Calculate float() using system locale set in LC_NUMERIC.'
    if not args:
        return 0.0

    return locale.atof(*args)


@vd.numericType('♯', fmtstr='%d')
class vlen(int):
    def __new__(cls, v=0):
        if isinstance(v, (vlen, int, float)):
            return super(vlen, cls).__new__(cls, v)
        else:
            return super(vlen, cls).__new__(cls, len(v))

    def __len__(self):
        return self

vd.addGlobals(anytype=anytype,
              vdtype=vdtype,
              deduceType=deduceType)



import pytest
import visidata

class TestCompleteExpr:
    def test_completer(self):
        vs = visidata.DirSheet('test', source=visidata.Path('.'))
        vs.reload()
        cexpr = visidata.CompleteExpr(vs)
        assert cexpr('fi', 0) == 'filename'  # visible column first
        assert cexpr('fi', 1) == 'filetype'  # hidden column second
        assert cexpr('logn', 0) == 'lognormvariate'  # global from math
        assert cexpr('a+logn', 0) == 'a+lognormvariate'

        assert cexpr('testv', 0) == 'testv'  # no match returns same

        visidata.vd.memoValue('testvalue', 42, '42')
        cexpr = visidata.CompleteExpr(vs)
        assert cexpr('testv', 0) == 'testvalue'



import os

from visidata import VisiData


@VisiData.api
def ansi(*args):
    os.write(1, b'\x1b'+b''.join([str(x).encode('utf-8') for x in args]))


@VisiData.api
def set_titlebar(vd, title:str):
    ansi(']2;', title, '\x07')



"""
Generate sparkline column for numeric columns
"""

from visidata import vd, Column, Sheet

__author__ = 'Lucas Messenger @layertwo'

vd.theme_option('disp_sparkline', '▁▂▃▄▅▆▇', 'characters to display sparkline')


def sparkline(*values):
    """
    From *values generate string sparkline
    """
    lines = vd.options.disp_sparkline
    values = [v for v in values if isinstance(v, (int, float))]
    mx = max(values)
    mn = min(values)
    w = (mx - mn) / len(lines)
    bounds = [(mn + w * i) for i in range(len(lines))]

    output = ''
    for val in values:
        for b in bounds:
            if mn == 0 and val == 0:
                output += ' '
                break
            if val < b:
                output += lines[bounds.index(b) - 1]
                break
        else:
            output += max(lines)
    return output


@Sheet.api
def addcol_sparkline(sheet, sourceCols):
    """
    Add sparkline column
    """
    c = Column('sparkline',
               sourceCols=sourceCols,
               getter=lambda c,r: sparkline(*tuple(c.getTypedValue(r) for c in c.sheet.sourceCols)))
    sheet.addColumn(c)


Sheet.addCommand(None, 'addcol-sparkline', 'addcol_sparkline(numericCols(nonKeyVisibleCols))', 'add sparkline of all numeric columns')
